import datetime
import logging
from operator import itemgetter
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from uuid import UUID

import click
from click_option_group import optgroup
from click_option_group import RequiredMutuallyExclusiveOptionGroup
from more_itertools import only
from mox_helpers.mox_util import ensure_class_in_lora
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.load_settings import load_settings
from tqdm import tqdm

import constants
from . import payloads
from .ad_logger import start_logging
from .ad_reader import ADParameterReader


LOG_FILE = "external_ad_users.log"

logger = logging.getLogger("ImportADGroup")


class ADMOImporter(object):
    def __init__(self):
        self.run_date = datetime.datetime.now().strftime("%Y-%m-%d")

        self.settings = self._get_settings()
        self.helper = self._get_mora_helper()

        self.root_ou_uuid = self.settings["integrations.ad.import_ou.mo_unit_uuid"]
        self.org_uuid = self.helper.read_organisation()

        self.AD_it_system_uuid = self._get_ad_it_system()

        self.ad_reader = self._get_ad_reader()

    def _get_settings(self) -> Dict[str, Any]:
        return load_settings()

    def _get_mora_helper(self) -> MoraHelper:
        return MoraHelper(hostname=self.settings["mora.base"], use_cache=False)

    def _get_ad_it_system(self) -> UUID:
        it_systems = self.helper.read_it_systems()
        ad_it_systems = only(
            filter(lambda x: x["name"] == constants.AD_it_system, it_systems)
        )
        return ad_it_systems["uuid"]

    def _get_ad_reader(self) -> ADParameterReader:
        ad_reader = ADParameterReader()
        ad_reader.cache_all(print_progress=True)
        return ad_reader

    def _ensure_class_in_lora(
        self, facet_bvn: str, class_value: str
    ) -> Tuple[UUID, bool]:
        return ensure_class_in_lora(facet_bvn, class_value)

    def _find_or_create_unit_and_classes(self):
        """
        Find uuids of the needed unit and classes for the import. If any unit or
        class is missing from MO, it will be created.The function returns a dict
        containg uuids needed to create users and engagements.
        """
        # TODO: Add a dynamic creation of classes
        job_type, _ = self._ensure_class_in_lora("engagement_job_function", "Ekstern")
        eng_type, _ = self._ensure_class_in_lora("engagement_type", "Ekstern")
        org_unit_type, _ = self._ensure_class_in_lora("org_unit_type", "Ekstern")

        unit = self.helper.read_ou(uuid=self.root_ou_uuid)
        if "status" in unit:  # Unit does not exist
            payload = payloads.create_unit(
                self.root_ou_uuid, "Eksterne Medarbejdere", org_unit_type, self.org_uuid
            )
            logger.debug("Create department payload: {}".format(payload))
            response = self.helper._mo_post("ou/create", payload)
            assert response.status_code == 201
            logger.info("Created unit for external employees")
            logger.debug("Response: {}".format(response.text))

        uuids = {
            "job_function": job_type,
            "engagement_type": eng_type,
            "unit_uuid": self.root_ou_uuid,
        }
        return uuids

    def _find_ou_users_in_ad(self) -> Dict[UUID, List]:
        """
        find users from AD that match a search string in DistinguishedName.
        """
        search_string = self.settings["integrations.ad.import_ou.search_string"]

        def filter_users(user: Dict) -> bool:
            name = user.get("DistinguishedName")
            if name:
                if search_string in name:
                    return True
            return False

        users = list(filter(filter_users, self.ad_reader.results.values()))
        uuids = map(itemgetter("ObjectGUID"), users)
        users_dict = dict(zip(uuids, users))
        return users_dict

    def _create_user(
        self, ad_user: Dict, cpr_field: str, uuid: Optional[str] = None
    ) -> Optional[UUID]:
        """
        Create or update a user in MO using an AD user as a template.
        The user will share uuid between MO and AD.
        :param ad_user: The ad_object to use as template for MO.
        :return: uuid of the the user.
        """
        cpr_raw = ad_user.get(cpr_field)
        if cpr_raw is None:
            return None
        cpr = cpr_raw.replace("-", "")

        payload = payloads.create_user(cpr, ad_user, self.org_uuid, uuid=uuid)
        logger.info("Create user payload: {}".format(payload))
        r = self.helper._mo_post("e/create", payload)
        assert r.status_code == 201
        user_uuid = UUID(r.json())
        logger.info("Created employee {}".format(user_uuid))
        return user_uuid

    def _connect_user_to_ad(self, mo_uuid: str, username: str) -> None:
        """Write user AD username to the AD it system"""

        logger.info("Connect user to AD: {}".format(username))

        payload = payloads.connect_it_system_to_user(
            str(mo_uuid), username, str(self.AD_it_system_uuid), from_date=self.run_date
        )
        logger.debug("AD account payload: {}".format(payload))
        response = self.helper._mo_post("details/create", payload)
        assert response.status_code == 201
        logger.debug("Added AD account info to {}".format(username))

    def _create_engagement(
        self, ad_user: Dict, uuids: Dict[str, UUID], mo_uuid: UUID = None
    ) -> None:
        """Create the engagement in MO"""
        # TODO: Check if we have start/end date of engagements in AD
        validity = {"from": self.run_date, "to": None}

        person_uuid = ad_user["ObjectGUID"]
        if mo_uuid:
            person_uuid = mo_uuid

        # TODO: Check if we can use job title from AD

        # Convert `uuid.UUID` values to strings before passing to `create_engagement`
        payload_kwargs = {k: str(v) for k, v in uuids.items()}
        payload = payloads.create_engagement(
            ad_user=ad_user,
            validity=validity,
            person_uuid=str(person_uuid),
            **payload_kwargs,
        )
        logger.info("Create engagement payload: {}".format(payload))
        response = self.helper._mo_post("details/create", payload)
        assert response.status_code == 201
        logger.info("Added engagement to {}".format(ad_user["SamAccountName"]))

    def cleanup_removed_users_from_mo(self) -> None:
        """Remove users in MO if they are no longer found as external users in AD."""
        yesterday = datetime.datetime.now() - datetime.timedelta(days=1)
        users = self._find_ou_users_in_ad()
        active_account_names = set(map(itemgetter("SamAccountName"), users.values()))

        mo_users = self.helper.read_organisation_people(self.root_ou_uuid)

        for mo_user_uuid, user in mo_users.items():
            AD_accounts = self.helper.get_e_itsystems(
                mo_user_uuid, it_system_uuid=self.AD_it_system_uuid
            )
            for account in AD_accounts:
                if account["user_key"] not in active_account_names:
                    # This users is in MO but not in AD:
                    logger.info(f"Terminating user with uuid={user['Person UUID']}")
                    payload = payloads.terminate_engagement(
                        user["Engagement UUID"], yesterday
                    )
                    logger.debug("Terminate payload: {}".format(payload))
                    response = self.helper._mo_post("details/terminate", payload)
                    logger.debug("Terminate response: {}".format(response.text))
                    response.raise_for_status()

                    # terminate username
                    payload["uuid"] = account["uuid"]
                    payload["type"] = "it"
                    logger.debug("Terminate it payload: {}".format(payload))
                    response = self.helper._mo_post("details/terminate", payload)
                    logger.debug("Terminate it response: {}".format(response.text))
                    response.raise_for_status()

    def create_or_update_users_in_mo(self) -> None:
        """
        Create users in MO that exist in AD but not in MO.
        Update name of users that has changed name in AD.
        """

        uuids = self._find_or_create_unit_and_classes()
        users = self._find_ou_users_in_ad()
        for AD in self.settings["integrations.ad"]:
            cpr_field = AD["cpr_field"]

            for user_uuid, ad_user in tqdm(
                users.items(), unit="Users", desc="Updating users"
            ):
                logger.info("Updating {}".format(ad_user["SamAccountName"]))
                cpr = ad_user[cpr_field]
                # Sometimes there is a temporary change of cpr in wich the
                # last character is replaced with an 'x'.
                # This user is ignored by the importer
                # until the cpr has been changed back.
                if cpr[-1].lower() == "x":
                    logger.info("Skipped due to 'x' in cpr.")
                    continue

                mo_user = self.helper.read_user(user_cpr=cpr)
                logger.info("Existing MO info: {}".format(mo_user))

                if mo_user:
                    mo_uuid = mo_user.get("uuid")
                else:
                    mo_uuid = self._create_user(ad_user, cpr_field)

                AD_username = self.helper.get_e_itsystems(
                    mo_uuid, self.AD_it_system_uuid
                )

                if not AD_username:
                    self._connect_user_to_ad(mo_uuid, ad_user["SamAccountName"])

                current_engagements = self.helper.read_user_engagement(user=mo_uuid)
                this_engagement = list(
                    filter(
                        lambda x: x.get("org_unit").get("uuid") == uuids["unit_uuid"],
                        current_engagements,
                    )
                )
                if not this_engagement:
                    self._create_engagement(ad_user, uuids, mo_uuid)


@click.command(help="AD->MO user import")
@optgroup.group("Action", cls=RequiredMutuallyExclusiveOptionGroup)
@optgroup.option("--create-or-update", is_flag=True)
@optgroup.option("--cleanup-removed-users", is_flag=True)
@optgroup.option("--full-sync", is_flag=True)
def import_ad_group(**args):
    """
    Command line interface for the AD to MO user import.
    """
    ad_import = ADMOImporter()

    if args.get("create_or_update"):
        ad_import.create_or_update_users_in_mo()

    if args.get("cleanup_removed_users"):
        ad_import.cleanup_removed_users_from_mo()

    if args.get("full_sync"):
        ad_import.create_or_update_users_in_mo()
        ad_import.cleanup_removed_users_from_mo()


if __name__ == "__main__":
    start_logging(LOG_FILE)
    import_ad_group()



import os

from os2mo_data_import import ImportHelper
from integrations.SD_Lon import sd_importer
from os2mo_data_import.caching_import import CachingImportUtility

MOX_BASE = os.environ.get('MOX_BASE', 'http://localhost:8080')
MORA_BASE = os.environ.get('MORA_BASE', 'http://localhost:5000')

# Import of Administration
adm_name = 'AdmOrg'

importer = ImportHelper(
    create_defaults=True,
    mox_base=MOX_BASE,
    mora_base=MORA_BASE,
    seperate_names=True,
    store_integration_data=False,
    ImportUtility=CachingImportUtility,
)

# importer.add_klasse(
#     identifier=adm_name,
#     facet_type_ref='org_unit_type',
#     user_key=adm_name,
#     scope='TEXT',
#     title=adm_name
# )
#
# importer.add_organisation_unit(
#     identifier=adm_name,
#     name=adm_name,
#     user_key=adm_name,
#     type_ref=adm_name,
#     date_from='1930-01-01',
#     date_to=None,
#     parent_ref=None
# )
#
# sd = sd_importer.SdImport(
#     importer,
#     org_only=True,
#     org_id_prefix='adm_'
# )
#
# sd.create_ou_tree(
#     create_orphan_container=False,
#     # 'Direktionen encoded with org prefix 'adm_'
#     sub_tree='fff9e2a6-d670-b656-c719-994eeac03a74',
#     super_unit = adm_name
# )


# Import of Lønorganisation
loen_name = 'LønOrg'

importer.add_klasse(
    identifier=loen_name,
    facet_type_ref='org_unit_type',
    user_key=loen_name,
    scope='TEXT',
    title=loen_name,
)

importer.add_organisation_unit(
    identifier=loen_name,
    name=loen_name,
    user_key=loen_name,
    type_ref=loen_name,
    date_from='1930-01-01',
    date_to=None,
    parent_ref=None,
)


sd = sd_importer.SdImport(
    importer,
    org_id_prefix=None,
)

sd.create_ou_tree(create_orphan_container=False, super_unit=loen_name)

sd.create_employees()

importer.import_all()



#!/usr/bin/env python3
#
# Copyright (c) 2017-2018, Magenta ApS
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import os
import csv
from os2mo_data_import import ImportHelper
from integrations import apos_importer


MUNICIPALTY_NAME = os.environ.get('MUNICIPALITY_NAME', 'APOS Import')
MOX_BASE = os.environ.get('MOX_BASE', 'http://localhost:5000')
MORA_BASE = os.environ.get('MORA_BASE', 'http://localhost:80')
MUNICIPALTY_CODE = os.environ.get('MUNICIPALITY_NAME', 0)

ean_file = '/opt/customer/dataimport/ballerup_udvalg/ean.csv'
ean_rows = {}
with open(ean_file) as csvfile:
    reader = csv.reader(csvfile, delimiter=';')
    for row in reader:
        ean_rows[row[0]] = row[1]


importer = ImportHelper(create_defaults=True,
                        mox_base=MOX_BASE,
                        mora_base=MORA_BASE,
                        system_name='APOS-Import',
                        end_marker='APOSSTOP',
                        store_integration_data=False,
                        seperate_names=True)


apos_import = apos_importer.AposImport(
    importer,
    MUNICIPALTY_NAME,
    MUNICIPALTY_CODE,
    org_uuid='00000000-0000-43b4-8415-1c70c2cd0cbd',
    ean=ean_rows
)

apos_import.create_facetter_and_klasser()

# Org træ
apos_import.create_ou_tree('b78993bb-d67f-405f-acc0-27653bd8c116')

# SD træ
apos_import.create_ou_tree('945bb286-9753-4f77-9082-a67a5d7bdbaf')

apos_import.create_managers_and_associatins()

apos_import.add_all_missing_employees()

importer.import_all()

"""
print('********************************')
print('Address challenges:')
for challenge in apos_import.address_challenges:
    print(apos_import.address_challenges[challenge])
print()

print('Address Errors:')
for error in apos_import.address_errors:
    print(apos_import.address_errors[error])
print()

print('Klassifikation Errors:')
for uuid, error in apos_import.klassifikation_errors.items():
    print(uuid)
print()

print('Duplicate people:')
for uuid, person in apos_import.duplicate_persons.items():
    print(person['@adresseringsnavn'])
"""



import random

from mimesis import Person

person_gen = Person("da")


def create_name():
    names = []
    names.append(person_gen.first_name())
    if random.random() > 0.25:
        names.append(person_gen.first_name())
    if random.random() > 0.8:
        names.append(person_gen.last_name())
    if random.random() > 0.99:
        names.append(person_gen.last_name())

    names.append(person_gen.last_name())
    return names


if __name__ == "__main__":
    import time

    t = time.time()
    for i in range(0, 1000):
        create_name()
    print(time.time() - t)



from enum import Enum
from typing import Optional

from pydantic import BaseModel


class JobFunction(str, Enum):
    job_position_identifier = "JobPositionIdentifier"
    employment_name = "EmploymentName"


class SDBasePerson(BaseModel):
    cpr: str
    given_name: Optional[str]
    surname: Optional[str]



import datetime
from abc import ABC
from abc import abstractmethod
from pathlib import Path
from typing import Dict
from typing import Optional

import click
import fs
from google.cloud import storage
from more_itertools import one
from ra_utils.load_settings import load_settings
from retrying import retry


class OpusReaderInterface(ABC):
    @abstractmethod
    def list_opus_files(self) -> Dict[datetime.datetime, str]:
        raise NotImplementedError()

    @abstractmethod
    def read_file(self, blob_name: str) -> str:
        raise NotImplementedError()

    def map_dates(self, dump_list) -> Dict[datetime.datetime, str]:
        """Transforms the list of opus_dumps to a dictionary with dates as keys"""
        dumps = {}
        for opus_dump in dump_list:
            date_part = opus_dump.name[4:18]
            export_time = datetime.datetime.strptime(date_part, "%Y%m%d%H%M%S")
            dumps[export_time] = opus_dump
        return dumps

    def read_latest(self):
        all_files = self.list_opus_files()
        latest_date = max(all_files.keys())
        return self.read_file(all_files[latest_date])


class GcloudOpusReader(OpusReaderInterface):
    def __init__(self, settings):
        settings = settings
        bucket_name = settings["integrations.opus.gcloud_bucket_name"]
        self.client = storage.Client()
        self.bucket = storage.Bucket(self.client, bucket_name)

    def list_opus_files(self) -> Dict[datetime.datetime, str]:
        all_files = list(self.client.list_blobs(self.bucket))
        for f in all_files:
            f.name = f.name.replace("production/", "")
        return self.map_dates(all_files)

    def read_file(self, blob) -> str:
        return blob.download_as_text()


retry_args = {
    "stop_max_attempt_number": 7,
    "wait_fixed": 2000,
}


class SMBOpusReader(OpusReaderInterface):
    @retry(**retry_args)
    def __init__(self, settings):
        self.settings = settings
        user = self.settings["integrations.opus.smb_user"]
        password = self.settings["integrations.opus.smb_password"]
        smb_host = self.settings["integrations.opus.smb_host"]
        self.smb_fs = fs.open_fs(f"smb://{user}:{password}@{smb_host}")

    @retry(**retry_args)
    def list_opus_files(self) -> Dict[datetime.datetime, str]:
        all_files = self.smb_fs.glob("*.xml")
        return self.map_dates(x.info for x in all_files)

    @retry(**retry_args)
    def read_file(self, glob) -> str:
        f = one(self.smb_fs.glob(glob.name))
        return self.smb_fs.readtext(f.path)


class LocalOpusReader(OpusReaderInterface):
    def __init__(self, settings):
        self.settings = settings

    def list_opus_files(self) -> Dict[datetime.datetime, str]:
        dump_path = Path(self.settings["integrations.opus.import.xml_path"])
        return self.map_dates(dump_path.glob("*.xml"))

    def read_file(self, filename) -> str:
        return filename.read_text()


def get_opus_filereader(settings: Optional[Dict] = None) -> OpusReaderInterface:
    """Get the correct opus reader interface based on values from settings."""
    settings = settings or load_settings()
    if settings.get("integrations.opus.gcloud_bucket_name"):
        return GcloudOpusReader(settings)
    if settings.get("integrations.opus.smb_host"):
        return SMBOpusReader(settings)
    return LocalOpusReader(settings)


@click.group()
def cli():
    """CLI for reading opus-files"""
    pass


@cli.command()
def read_last():
    """Read latest opus-file"""
    ofr = get_opus_filereader()
    click.echo(ofr.read_latest())


@cli.command()
def list_files():
    """Show dates of all opus-files"""
    ofr = get_opus_filereader()
    dumps = ofr.list_opus_files()
    dates = sorted(dumps.keys())
    for date in dates:
        click.echo(date)


@cli.command()
@click.option("--date", type=click.DateTime())
def read_file(date):
    """Read opus-file from specific date. If no date is supplied show all available dates"""
    ofr = get_opus_filereader()
    dumps = ofr.list_opus_files()

    if not date:
        dates = sorted(dumps.keys())
        for date in dates:
            click.echo(date)
        click.echo(f'Choose from above and provide as parameter, eg. --date="{date}"')
    else:
        click.echo(ofr.read_file(dumps[date]))


if __name__ == "__main__":
    cli()



import tempfile
import unittest

from hypothesis import given
from hypothesis.strategies import text

from integrations.opus.opus_file_reader import get_opus_filereader
from integrations.opus.opus_file_reader import LocalOpusReader


class OpusFileReader_test(unittest.TestCase):
    """
    Tests for opus file reader interface
    """

    def test_local(self):
        settings = {}
        ofr = get_opus_filereader(settings=settings)
        assert isinstance(ofr, LocalOpusReader)

    def test_empty_path(self):
        with tempfile.TemporaryDirectory() as tmppath:
            settings = {"integrations.opus.import.xml_path": tmppath}

            ofr = get_opus_filereader(settings=settings)
            assert len(ofr.list_opus_files()) == 0

    @given(text())
    def test_text_in_file(self, dummy_text):
        dummy_text = repr(dummy_text)
        filename = "/ZLPB20100101145823.xml"
        with tempfile.TemporaryDirectory() as tmppath:
            with open(tmppath + filename, "w") as opus_file:
                opus_file.write(dummy_text)

            settings = {"integrations.opus.import.xml_path": tmppath}
            ofr = get_opus_filereader(settings=settings)
            files_list = ofr.list_opus_files()
            assert len(files_list) == 1
            text_in = ofr.read_latest()
            assert text_in == dummy_text


if __name__ == "__main__":
    unittest.main()



import datetime
import logging
from typing import Tuple

import click
from alembic.migration import MigrationContext
from alembic.operations import Operations
from ra_utils.load_settings import load_settings
from sqlalchemy import create_engine
from sqlalchemy import Index
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session
from sqlalchemy.orm import sessionmaker
from tqdm import tqdm

from exporters.sql_export.lora_cache import LoraCache
from exporters.sql_export.sql_table_defs import Adresse
from exporters.sql_export.sql_table_defs import Base
from exporters.sql_export.sql_table_defs import Bruger
from exporters.sql_export.sql_table_defs import DARAdresse
from exporters.sql_export.sql_table_defs import Engagement
from exporters.sql_export.sql_table_defs import Enhed
from exporters.sql_export.sql_table_defs import Enhedssammenkobling
from exporters.sql_export.sql_table_defs import Facet
from exporters.sql_export.sql_table_defs import ItForbindelse
from exporters.sql_export.sql_table_defs import ItSystem
from exporters.sql_export.sql_table_defs import Klasse
from exporters.sql_export.sql_table_defs import KLE
from exporters.sql_export.sql_table_defs import Kvittering
from exporters.sql_export.sql_table_defs import Leder
from exporters.sql_export.sql_table_defs import LederAnsvar
from exporters.sql_export.sql_table_defs import Orlov
from exporters.sql_export.sql_table_defs import Rolle
from exporters.sql_export.sql_table_defs import Tilknytning
from exporters.sql_export.sql_url import DatabaseFunction
from exporters.sql_export.sql_url import generate_connection_url
from exporters.sql_export.sql_url import generate_engine_settings


LOG_LEVEL = logging.DEBUG
LOG_FILE = "sql_export.log"

logger = logging.getLogger("SqlExport")


class SqlExport:
    def __init__(self, force_sqlite=False, historic=False, settings=None):
        logger.info("Start SQL export")
        self.force_sqlite = force_sqlite
        self.historic = historic
        self.settings = settings
        self.engine = self._get_engine()
        self.export_cpr = self._get_export_cpr_setting()

    def _get_engine(self) -> Engine:
        database_function = DatabaseFunction.ACTUAL_STATE
        if self.historic:
            database_function = DatabaseFunction.ACTUAL_STATE_HISTORIC
        db_string = generate_connection_url(
            database_function, force_sqlite=self.force_sqlite, settings=self.settings
        )
        engine_settings = generate_engine_settings(
            database_function, force_sqlite=self.force_sqlite, settings=self.settings
        )
        return create_engine(db_string, **engine_settings)

    def _get_export_cpr_setting(self) -> bool:
        return self.settings.get("exporters.actual_state.export_cpr", True)

    def _get_lora_cache(self, resolve_dar, use_pickle) -> LoraCache:
        if self.historic:
            lc = LoraCache(resolve_dar=resolve_dar, full_history=True)
            lc.populate_cache(dry_run=use_pickle)
        else:
            lc = LoraCache(resolve_dar=resolve_dar)
            lc.populate_cache(dry_run=use_pickle)
            lc.calculate_derived_unit_data()
            lc.calculate_primary_engagements()
        return lc

    def _get_db_session(self) -> Session:
        Session = sessionmaker(bind=self.engine, autoflush=False)
        return Session()

    def _get_lora_class(self, uuid: str) -> Tuple[str, dict]:
        cls: dict = self.lc.classes.get(uuid) or {"title": uuid}
        return uuid, cls

    def perform_export(self, resolve_dar=True, use_pickle=None):
        def timestamp():
            return datetime.datetime.now()

        trunc_tables = dict(Base.metadata.tables)
        trunc_tables.pop("kvittering")

        Base.metadata.drop_all(self.engine, tables=trunc_tables.values())
        Base.metadata.create_all(self.engine)

        self.session = self._get_db_session()

        query_time = timestamp()
        kvittering = self._add_receipt(query_time)
        self.lc = self._get_lora_cache(resolve_dar, use_pickle)

        start_delivery_time = timestamp()
        self._update_receipt(kvittering, start_delivery_time)

        tasks = [
            self._add_classification,
            self._add_users_and_units,
            self._add_addresses,
            self._add_dar_addresses,
            self._add_engagements,
            self._add_associations,
            self._add_managers,
            self._add_it_systems,
            self._add_kles,
            self._add_related,
        ]
        for task in tqdm(tasks, desc="SQLExport", unit="task"):
            task()

        end_delivery_time = timestamp()
        self._update_receipt(kvittering, start_delivery_time, end_delivery_time)

    def swap_tables(self):
        """Swap tables around to present the exported data.

        Swaps the current tables to old tables, then swaps write tables to current.
        Finally drops the old tables leaving just the current tables.
        """
        connection = self.engine.connect()
        ctx = MigrationContext.configure(connection)
        op = Operations(ctx)

        def gen_table_names(write_table):
            """Generate current and old table names from write tables."""
            # Current tables do not have the prefix 'w'
            current_table = write_table[1:]
            old_table = current_table + "_old"
            return write_table, current_table, old_table

        tables = dict(Base.metadata.tables)
        tables.pop("kvittering")
        tables = tables.keys()
        tables = list(map(gen_table_names, tables))

        # Drop any left-over old tables that may exist
        with ctx.begin_transaction():
            for _, _, old_table in tables:
                try:
                    op.drop_table(old_table)
                except Exception:
                    pass

        # Rename current to old and write to current
        with ctx.begin_transaction():
            for write_table, current_table, old_table in tables:
                # Rename current table to old table
                # No current tables is OK
                try:
                    op.rename_table(current_table, old_table)
                except Exception:
                    pass
                # Rename write table to current table
                op.rename_table(write_table, current_table)

        # Drop any old tables that may exist
        with ctx.begin_transaction():
            for _, _, old_table in tables:
                # Drop old tables
                try:
                    op.drop_table(old_table)
                except Exception:
                    pass

    def _add_classification(self, output=False):
        logger.info("Add classification")
        for facet, facet_info in tqdm(
            self.lc.facets.items(), desc="Export facet", unit="facet"
        ):
            sql_facet = Facet(
                uuid=facet,
                bvn=facet_info["user_key"],
            )
            self.session.add(sql_facet)
        self.session.commit()

        for klasse, klasse_info in tqdm(
            self.lc.classes.items(), desc="Export class", unit="class"
        ):
            sql_class = Klasse(
                uuid=klasse,
                bvn=klasse_info["user_key"],
                titel=klasse_info["title"],
                facet_uuid=klasse_info["facet"],
                facet_bvn=self.lc.facets[klasse_info["facet"]]["user_key"],
            )
            self.session.add(sql_class)
        self.session.commit()

        if output:
            for result in self.engine.execute("select * from facetter limit 4"):
                print(result.items())
            for result in self.engine.execute("select * from klasser limit 4"):
                print(result.items())

    def _add_users_and_units(self, output=False):
        logger.info("Add users and units")
        for user, user_effects in tqdm(
            self.lc.users.items(), desc="Export user", unit="user"
        ):
            for user_info in user_effects:
                sql_user = Bruger(
                    uuid=user,
                    bvn=user_info["user_key"],
                    fornavn=user_info["fornavn"],
                    efternavn=user_info["efternavn"],
                    kaldenavn_fornavn=user_info["kaldenavn_fornavn"],
                    kaldenavn_efternavn=user_info["kaldenavn_efternavn"],
                    cpr=user_info["cpr"] if self.export_cpr else "",
                    startdato=user_info["from_date"],
                    slutdato=user_info["to_date"],
                )
                self.session.add(sql_user)
            self.session.commit()

        for unit, unit_validities in tqdm(
            self.lc.units.items(), desc="Export unit", unit="unit"
        ):
            for unit_info in unit_validities:
                location = unit_info.get("location")
                manager_uuid = unit_info.get("manager_uuid")
                acting_manager_uuid = unit_info.get("acting_manager_uuid")

                unit_type = unit_info["unit_type"]
                enhedsniveau_titel = ""
                if unit_info["level"]:
                    enhedsniveau_titel = self.lc.classes[unit_info["level"]]["title"]
                sql_unit = Enhed(
                    uuid=unit,
                    navn=unit_info["name"],
                    bvn=unit_info["user_key"],
                    forældreenhed_uuid=unit_info["parent"],
                    enhedstype_uuid=unit_type,
                    enhedsniveau_uuid=unit_info["level"],
                    organisatorisk_sti=location,
                    leder_uuid=manager_uuid,
                    fungerende_leder_uuid=acting_manager_uuid,
                    enhedstype_titel=self.lc.classes[unit_type]["title"],
                    enhedsniveau_titel=enhedsniveau_titel,
                    startdato=unit_info["from_date"],
                    slutdato=unit_info["to_date"],
                )
                self.session.add(sql_unit)
            self.session.commit()

        # create supplementary index for quick toplevel lookup
        # when rewriting whole table this is quicker than maintaining
        # the index for every row inserted
        organisatorisk_sti_index = Index(
            "organisatorisk_sti_index", Enhed.organisatorisk_sti
        )
        organisatorisk_sti_index.create(bind=self.engine)

        if output:
            for result in self.engine.execute("select * from brugere limit 5"):
                print(result)
            for result in self.engine.execute("select * from enheder limit 5"):
                print(result)

    def _add_engagements(self, output=False):
        logger.info("Add engagements")
        for engagement, engagement_validity in tqdm(
            self.lc.engagements.items(), desc="Export engagement", unit="engagement"
        ):
            for engagement_info in engagement_validity:
                if engagement_info["primary_type"] is not None:
                    primærtype_titel = self.lc.classes[engagement_info["primary_type"]][
                        "title"
                    ]
                else:
                    primærtype_titel = ""

                engagement_type_uuid = engagement_info["engagement_type"]
                job_function_uuid, job_function_class = self._get_lora_class(
                    engagement_info["job_function"]
                )

                sql_engagement = Engagement(
                    uuid=engagement,
                    enhed_uuid=engagement_info["unit"],
                    bruger_uuid=engagement_info["user"],
                    bvn=engagement_info["user_key"],
                    engagementstype_uuid=engagement_info["engagement_type"],
                    primær_boolean=engagement_info.get("primary_boolean"),
                    arbejdstidsfraktion=engagement_info["fraction"],
                    engagementstype_titel=self.lc.classes.get(
                        engagement_type_uuid, {"title": engagement_type_uuid}
                    )["title"],
                    primærtype_titel=primærtype_titel,
                    stillingsbetegnelse_uuid=job_function_uuid,
                    stillingsbetegnelse_titel=job_function_class["title"],
                    primærtype_uuid=engagement_info["primary_type"],
                    startdato=engagement_info["from_date"],
                    slutdato=engagement_info["to_date"],
                    **engagement_info["extensions"]
                )
                self.session.add(sql_engagement)
            self.session.commit()

        if output:
            for result in self.engine.execute("select * from engagementer limit 5"):
                print(result.items())

    def _add_addresses(self, output=False):
        logger.info("Add addresses")
        for address, address_validities in tqdm(
            self.lc.addresses.items(), desc="Export address", unit="address"
        ):
            for address_info in address_validities:
                visibility_text = None
                if address_info["visibility"] is not None:
                    visibility_text = self.lc.classes[address_info["visibility"]][
                        "title"
                    ]
                visibility_scope = None
                if address_info["visibility"] is not None:
                    visibility_scope = self.lc.classes[address_info["visibility"]][
                        "scope"
                    ]

                sql_address = Adresse(
                    uuid=address,
                    enhed_uuid=address_info["unit"],
                    bruger_uuid=address_info["user"],
                    værdi=address_info["value"],
                    dar_uuid=address_info["dar_uuid"],
                    adressetype_uuid=address_info["adresse_type"],
                    adressetype_bvn=self.lc.classes[address_info["adresse_type"]][
                        "user_key"
                    ],
                    adressetype_scope=address_info["scope"],
                    adressetype_titel=self.lc.classes[address_info["adresse_type"]][
                        "title"
                    ],
                    synlighed_uuid=address_info["visibility"],
                    synlighed_scope=visibility_scope,
                    synlighed_titel=visibility_text,
                    startdato=address_info["from_date"],
                    slutdato=address_info["to_date"],
                )
                self.session.add(sql_address)
            self.session.commit()
        if output:
            for result in self.engine.execute("select * from adresser limit 10"):
                print(result.items())

    def _add_dar_addresses(self, output=False):
        logger.info("Add DAR addresses")
        for address, address_info in tqdm(
            self.lc.dar_cache.items(), desc="Export DAR", unit="DAR"
        ):
            sql_address = DARAdresse(
                uuid=address,
                **{
                    key: value
                    for key, value in address_info.items()
                    if key in DARAdresse.__table__.columns.keys() and key != "id"
                }
            )
            self.session.add(sql_address)
        self.session.commit()
        if output:
            for result in self.engine.execute("select * from dar_adresser limit 10"):
                print(result.items())

    def _add_associations(self, output=False):
        logger.info("Add associations")
        for association, association_validity in tqdm(
            self.lc.associations.items(), desc="Export association", unit="association"
        ):
            for association_info in association_validity:
                association_type_uuid, association_type_class = self._get_lora_class(
                    association_info["association_type"]
                )
                job_function_uuid, job_function_class = self._get_lora_class(
                    association_info["job_function"]
                )
                sql_association = Tilknytning(
                    uuid=association,
                    bruger_uuid=association_info["user"],
                    enhed_uuid=association_info["unit"],
                    bvn=association_info["user_key"],
                    tilknytningstype_uuid=association_type_uuid,
                    tilknytningstype_titel=association_type_class["title"],
                    startdato=association_info["from_date"],
                    slutdato=association_info["to_date"],
                    it_forbindelse_uuid=association_info["it_user"],
                    stillingsbetegnelse_uuid=job_function_uuid,
                    stillingsbetegnelse_titel=job_function_class["title"],
                    primær_boolean=association_info.get("primary_boolean"),
                )
                self.session.add(sql_association)
            self.session.commit()

        for role, role_validity in tqdm(
            self.lc.roles.items(), desc="Export role", unit="role"
        ):
            for role_info in role_validity:
                sql_role = Rolle(
                    uuid=role,
                    bruger_uuid=role_info["user"],
                    enhed_uuid=role_info["unit"],
                    rolletype_uuid=role_info["role_type"],
                    rolletype_titel=self.lc.classes[role_info["role_type"]]["title"],
                    startdato=role_info["from_date"],
                    slutdato=role_info["to_date"],
                )
                self.session.add(sql_role)
            self.session.commit()

        for leave, leave_validity in tqdm(
            self.lc.leaves.items(), desc="Export leave", unit="leave"
        ):
            for leave_info in leave_validity:
                leave_type = leave_info["leave_type"]
                sql_leave = Orlov(
                    uuid=leave,
                    bvn=leave_info["user_key"],
                    bruger_uuid=leave_info["user"],
                    orlovstype_uuid=leave_type,
                    orlovstype_titel=self.lc.classes[leave_type]["title"],
                    engagement_uuid=leave_info["engagement"],
                    startdato=leave_info["from_date"],
                    slutdato=leave_info["to_date"],
                )
                self.session.add(sql_leave)
            self.session.commit()
        if output:
            for result in self.engine.execute("select * from tilknytninger limit 4"):
                print(result.items())
            for result in self.engine.execute("select * from orlover limit 4"):
                print(result.items())
            for result in self.engine.execute("select * from roller limit 4"):
                print(result.items())

    def _add_it_systems(self, output=False):
        logger.info("Add IT systems")
        for itsystem, itsystem_info in tqdm(
            self.lc.itsystems.items(), desc="Export itsystem", unit="itsystem"
        ):
            sql_itsystem = ItSystem(uuid=itsystem, navn=itsystem_info["name"])
            self.session.add(sql_itsystem)
        self.session.commit()

        for it_connection, it_connection_validity in tqdm(
            self.lc.it_connections.items(),
            desc="Export it connection",
            unit="it connection",
        ):
            for it_connection_info in it_connection_validity:
                sql_it_connection = ItForbindelse(
                    uuid=it_connection,
                    it_system_uuid=it_connection_info["itsystem"],
                    bruger_uuid=it_connection_info["user"],
                    enhed_uuid=it_connection_info["unit"],
                    brugernavn=it_connection_info["username"],
                    startdato=it_connection_info["from_date"],
                    slutdato=it_connection_info["to_date"],
                    primær_boolean=it_connection_info["primary_type"] is not None,
                )
                self.session.add(sql_it_connection)
            self.session.commit()
        if output:
            for result in self.engine.execute("select * from it_systemer limit 2"):
                print(result.items())

            for result in self.engine.execute("select * from it_forbindelser limit 2"):
                print(result.items())

    def _add_kles(self, output=False):
        logger.info("Add KLES")
        for kle, kle_validity in tqdm(
            self.lc.kles.items(), desc="Export KLE", unit="KLE"
        ):
            for kle_info in kle_validity:
                sql_kle = KLE(
                    uuid=kle,
                    enhed_uuid=kle_info["unit"],
                    kle_aspekt_uuid=kle_info["kle_aspect"],
                    kle_aspekt_titel=self.lc.classes[kle_info["kle_aspect"]]["title"],
                    kle_nummer_uuid=kle_info["kle_number"],
                    kle_nummer_titel=self.lc.classes[kle_info["kle_number"]]["title"],
                    startdato=kle_info["from_date"],
                    slutdato=kle_info["to_date"],
                )
                self.session.add(sql_kle)
            self.session.commit()
        if output:
            for result in self.engine.execute("select * from kle limit 10"):
                print(result.items())

    def _add_receipt(self, query_time, start_time=None, end_time=None, output=False):
        logger.info("Add Receipt")
        sql_kvittering = Kvittering(
            query_tid=query_time,
            start_levering_tid=start_time,
            slut_levering_tid=end_time,
        )
        self.session.add(sql_kvittering)
        self.session.commit()
        if output:
            for result in self.engine.execute("select * from kvittering limit 10"):
                print(result.items())
        return sql_kvittering

    def _update_receipt(
        self, sql_kvittering, start_time=None, end_time=None, output=False
    ):
        logger.info("Update Receipt")
        sql_kvittering.start_levering_tid = start_time
        sql_kvittering.slut_levering_tid = end_time
        self.session.commit()
        if output:
            for result in self.engine.execute("select * from kvittering limit 10"):
                print(result.items())

    def _add_related(self, output=False):
        logger.info("Add Enhedssammenkobling")
        for related, related_validity in tqdm(
            self.lc.related.items(), desc="Export related", unit="related"
        ):
            for related_info in related_validity:
                sql_related = Enhedssammenkobling(
                    uuid=related,
                    enhed1_uuid=related_info["unit1_uuid"],
                    enhed2_uuid=related_info["unit2_uuid"],
                    startdato=related_info["from_date"],
                    slutdato=related_info["to_date"],
                )
                self.session.add(sql_related)
            self.session.commit()
        if output:
            for result in self.engine.execute(
                "select * from enhedssammenkobling limit 10"
            ):
                print(result.items())

    def _add_managers(self, output=False):
        logger.info("Add managers")
        for manager, manager_validity in tqdm(
            self.lc.managers.items(), desc="Export manager", unit="manager"
        ):
            for manager_info in manager_validity:
                sql_manager = Leder(
                    uuid=manager,
                    bruger_uuid=manager_info["user"],
                    enhed_uuid=manager_info["unit"],
                    niveautype_uuid=manager_info["manager_level"],
                    ledertype_uuid=manager_info["manager_type"],
                    niveautype_titel=self.lc.classes[manager_info["manager_level"]][
                        "title"
                    ],
                    ledertype_titel=self.lc.classes[manager_info["manager_type"]][
                        "title"
                    ],
                    startdato=manager_info["from_date"],
                    slutdato=manager_info["to_date"],
                )
                self.session.add(sql_manager)

                for responsibility in manager_info["manager_responsibility"]:
                    sql_responsibility = LederAnsvar(
                        leder_uuid=manager,
                        lederansvar_uuid=responsibility,
                        lederansvar_titel=self.lc.classes[responsibility]["title"],
                        startdato=manager_info["from_date"],
                        slutdato=manager_info["to_date"],
                    )
                    self.session.add(sql_responsibility)
            self.session.commit()
        if output:
            for result in self.engine.execute("select * from ledere limit 10"):
                print(result.items())
            for result in self.engine.execute("select * from leder_ansvar limit 10"):
                print(result.items())


@click.command(help="SQL export")
@click.option("--resolve-dar", is_flag=True)
@click.option("--historic", is_flag=True)
@click.option("--read-from-cache", is_flag=True, envvar="USE_CACHED_LORACACHE")
@click.option("--force-sqlite", is_flag=True)
def cli(**args):
    """
    Command line interface.
    """
    logger.info("Command line args: %r", args)

    settings = load_settings()

    sql_export = SqlExport(
        force_sqlite=args["force_sqlite"],
        historic=args["historic"],
        settings=settings,
    )
    sql_export.perform_export(
        resolve_dar=args["resolve_dar"],
        use_pickle=args["read_from_cache"],
    )
    sql_export.swap_tables()
    logger.info("*SQL export ended*")


if __name__ == "__main__":

    for name in logging.root.manager.loggerDict:
        if name in ("LoraCache", "SqlExport"):
            logging.getLogger(name).setLevel(LOG_LEVEL)
        else:
            logging.getLogger(name).setLevel(logging.ERROR)

    logging.basicConfig(
        format="%(levelname)s %(asctime)s %(name)s %(message)s",
        level=LOG_LEVEL,
        filename=LOG_FILE,
    )

    cli()



from more_itertools import first


def choose_public_address_helper(
    candidates, prioritized_classes, scope_getter, address_type_getter, uuid_getter
):
    """Pick the most desirable valid candidate address.

    An address candidate is considered valid if its visibility is PUBLIC or UNSET.

    An address candidates desirability is inversely propertional to its address
    types position inside prioritized_classes. I.e. address candidates whose address
    types occur earlier in prioritized_classes are more desirable than address
    candidates whose address types occur later (or yet worse, not at all).
    Additionally lower UUIDs are considered more desirable than larger ones, this is
    to ensure a consistent order regardless of the order of the candidates.

    Args:
        candidates: List of Address entries.
        prioritized_classes: List of Address Type UUIDs.
        scope_getter: Function from address entry to visibility scope.
        address_type_getter: Function from address entry to address type uuid.
        uuid_getter: Function from address entry to address entry uuid.

    Returns:
        Address entry: The address entry that is the most desirable.
    """

    def filter_by_visibility(candidate):
        """Predicate for filtering on visibility.

        Args:
            candidate: Address entry.

        Returns:
            bool: True for candidates with PUBLIC or UNSET visibility.
                  False otherwise.
        """
        visibility_scope = scope_getter(candidate)
        return visibility_scope is None or visibility_scope == "PUBLIC"

    def determine_candidate_desirability(candidate):
        """Predicate for determining desirability of an address candidate.

        The lower the value returned, the more desirable the candidate is.

        Args:
            candidate: Address entry.

        Returns:
            int: Index of the candidates address_type inside prioritized_classes.
                 Length of prioritized_classes if no match is found.
        """
        address_type_uuid = address_type_getter(candidate)
        try:
            priority = prioritized_classes.index(address_type_uuid)
        except ValueError:
            priority = len(prioritized_classes)
        return priority, uuid_getter(candidate)

    # Filter candidates to only keep valid ones
    candidates = filter(filter_by_visibility, candidates)

    # If no prioritized_classes are provided, all the entries are equally desirable.
    # Thus we can just return the entry with the lowest uuid.
    if not prioritized_classes:
        return min(candidates, key=uuid_getter, default=None)

    # If prioritized_classes are provided, we want to return the most desirable one.
    # The lowest index is the most desirable.
    return min(candidates, key=determine_candidate_desirability, default=None)


def mora_choose_public_address(candidates, prioritized_classes):
    """See choose_public_address_helper.

    Candidates are a list of MO address entries.
    """

    def scope_getter(candidate):
        if "visibility" not in candidate:
            return None
        if candidate["visibility"] is None:
            return None
        return candidate["visibility"]["scope"]

    def address_type_getter(candidate):
        return candidate["address_type"]["uuid"]

    def uuid_getter(candidate):
        return candidate["uuid"]

    return choose_public_address_helper(
        candidates,
        prioritized_classes,
        scope_getter,
        address_type_getter,
        uuid_getter,
    )


def lc_choose_public_address(candidates, prioritized_classes, lc):
    """See choose_public_address_helper.

    Candidates are a list of LoraCache address entries.
    """

    def scope_getter(candidate):
        if "visibility" not in candidate:
            return None
        if candidate["visibility"] is None:
            return None
        return lc.classes[candidate["visibility"]]["scope"]

    def address_type_getter(candidate):
        return candidate["adresse_type"]

    def uuid_getter(candidate):
        return candidate["uuid"]

    return choose_public_address_helper(
        candidates,
        prioritized_classes,
        scope_getter,
        address_type_getter,
        uuid_getter,
    )


def lcdb_choose_public_address(candidates, prioritized_classes):
    """See choose_public_address_helper.

    Candidates are a list of LoraCache sqlalchemy address entries.
    """

    def scope_getter(candidate):
        scope = candidate.synlighed_scope
        return scope or None

    def address_type_getter(candidate):
        return candidate.adressetype_uuid

    def uuid_getter(candidate):
        return candidate.uuid

    return choose_public_address_helper(
        candidates,
        prioritized_classes,
        scope_getter,
        address_type_getter,
        uuid_getter,
    )


def choose_public_address(candidates, prioritized_classes):
    """See mora_choose_public_address."""
    return mora_choose_public_address(candidates, prioritized_classes)



from ra_utils.async_to_sync import async_to_sync

from constants import FK_org_uuid_it_system
from integrations.fkorg_it_systems.import_fkorg_orgunits import check_it_system_value


class MockResponse:
    def __init__(self, return_value):
        self.return_value = return_value

    def json(self):
        return self.return_value


class MockClient:
    def __init__(self, return_value):
        self.return_value = return_value

    async def get(self, url):
        return MockResponse(self.return_value)


@async_to_sync
async def test_no_it_system():
    moModel = MockClient([])
    uuid, changed = await check_it_system_value(moModel, "dummy_org", "new")
    assert not uuid
    assert changed


@async_to_sync
async def test_it_system_changed():
    account_uuid = "account_uuid"
    moModel = MockClient(
        [
            {
                "uuid": account_uuid,
                "user_key": "old",
                "itsystem": {"name": FK_org_uuid_it_system},
            }
        ]
    )
    uuid, changed = await check_it_system_value(moModel, "dummy_org", "new")
    assert uuid == account_uuid
    assert changed


@async_to_sync
async def test_it_system_unchanged():
    account_uuid = "account_uuid"
    moModel = MockClient(
        [
            {
                "uuid": account_uuid,
                "user_key": "old",
                "itsystem": {"name": FK_org_uuid_it_system},
            }
        ]
    )
    uuid, changed = await check_it_system_value(moModel, "dummy_org", "old")
    assert uuid == account_uuid
    assert not changed



from typing import Any
from typing import Dict
from typing import Optional
from uuid import UUID

import uuids
from pydantic import BaseSettings
from ra_utils.load_settings import load_settings


def json_config_settings_source(settings: BaseSettings) -> Dict[str, Any]:
    settings_json = load_settings() or {}
    prefix = "integrations.aarhus_los"

    def _get_setting_value(key: str, default: Any = None):
        return settings_json.get(f"{prefix}.{key}", default)

    return dict(
        ftp_url=_get_setting_value("ftp_url", "ftp.aarhuskommune.dk"),
        ftp_user=_get_setting_value("ftp_user"),
        ftp_pass=_get_setting_value("ftp_pass"),
        ftp_folder=_get_setting_value("ftp_folder", "TEST"),
        import_state_file=_get_setting_value("state_file"),
        import_csv_folder=_get_setting_value("import_csv_folder"),
        azid_it_system_uuid=_get_setting_value(
            "azid_it_system_uuid", uuids.AZID_SYSTEM
        ),
    )


class ImproperlyConfigured(Exception):
    pass


class Settings(BaseSettings):
    ftp_url: str
    ftp_user: str
    ftp_pass: str
    ftp_folder: str
    import_state_file: str
    import_csv_folder: Optional[str]
    mox_base: str = "http://localhost:8080"
    mora_base: str = "http://localhost:5000"
    queries_dir: str = "/opt/docker/os2mo/queries"
    max_concurrent_requests: int = 4
    os2mo_chunk_size: int = 20
    azid_it_system_uuid: UUID

    class Config:
        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                init_settings,
                env_settings,
                json_config_settings_source,
                file_secret_settings,
            )

    @classmethod
    def from_kwargs(cls, **kwargs):
        """Return a `Settings` instance populated based on `kwargs`.

        This is usually called from a function decorated with `@click.command`:

            @click.command()
            def main(**kwargs):
                settings = Settings.from_kwargs(**kwargs)
        """
        instance = cls(**kwargs)
        # Save populated `Settings` instance so `get_config` can retrieve it
        cls._instance = instance
        return instance


def get_config() -> Settings:
    try:
        return Settings._instance  # type: ignore
    except AttributeError:
        raise ImproperlyConfigured(
            "Settings are not configured - did you call `Settings.from_kwargs`?"
        )



# Script to import a mapping of MO-uuids to FK-org uuids as it-accounts on org_units in OS2MO
# metacli import_fkorg_orgunits integrations/fk-org\ it\ systems/test_uuids.json --dry-run
import json
from datetime import datetime
from typing import Optional
from typing import Tuple

import click
from more_itertools import one
from more_itertools import only
from ra_utils.async_to_sync import async_to_sync
from raclients.modelclient.mo import ModelClient
from ramodels.mo.details import ITUser

import constants


class DryRunClient(ModelClient):
    async def upload(self, *args, **kwargs):
        click.echo(f"dry-run. Would create {len(args[0])} new accounts")

    async def edit(self, *args, **kwargs):
        click.echo(f"dry-run. Would edit {len(args[0])} accounts")


async def check_it_system_value(
    moclient: ModelClient, mo_uuid: str, fk_org_uuid: str
) -> Tuple[Optional[str], bool]:
    """Checks the fk_org it accounts of an org_unit
    Returns the (optional) uuid of the current fk_org_uuid it system account
    and a boolean which tells whether or not the account should be changed or created.
    """
    current_it_accounts = await moclient.get(f"/service/ou/{mo_uuid}/details/it")
    fk_org_account = only(
        filter(
            lambda it: it["itsystem"]["name"] == constants.FK_org_uuid_it_system,
            current_it_accounts.json(),
        )
    )
    if fk_org_account is None:
        account_uuid, change = None, True
    else:
        account_uuid = fk_org_account["uuid"]
        change = True if fk_org_account["user_key"] != fk_org_uuid else False
    return account_uuid, change


@click.command()
@click.argument("input-file")
@click.option("--mora-base", envvar="BASE_URL")
@click.option("--client-id", envvar="CLIENT_ID")
@click.option(
    "--client-secret",
    envvar="CLIENT_SECRET",
)
@click.option("--auth-server", envvar="AUTH_SERVER")
@click.option("--dry-run", is_flag=True)
@async_to_sync
async def cli(
    input_file, mora_base, client_id, client_secret, auth_server, dry_run
) -> None:
    # Read the file
    with open(input_file) as f:
        import_rows = json.load(f)
    # Setup MO client
    MoModelClient = DryRunClient if dry_run else ModelClient

    mo_model_client = MoModelClient(
        base_url=mora_base,
        client_secret=client_secret,
        client_id=client_id,
        auth_server=auth_server,
        auth_realm="mo",
    )
    # Find uuid of the fk-org it-system
    root = await mo_model_client.get("/service/o/")
    root_uuid = one(root.json())["uuid"]
    it_systems = await mo_model_client.get(f"/service/o/{root_uuid}/it/")
    fk_org_it = one(
        filter(
            lambda it: it["name"] == constants.FK_org_uuid_it_system, it_systems.json()
        )
    )
    edits = []
    creates = []
    today = datetime.utcnow().strftime("%Y-%m-%d")

    for row in import_rows:
        account_uuid, changed = await check_it_system_value(
            mo_model_client, row["mo_uuid"], row["fk_org_uuid"]
        )
        payload = {
            "user_key": row["fk_org_uuid"],
            "itsystem_uuid": fk_org_it["uuid"],
            "org_unit_uuid": row["mo_uuid"],
            "from_date": today,
        }
        if account_uuid is None:
            creates.append(ITUser.from_simplified_fields(**payload))
        elif changed:
            payload["uuid"] = account_uuid
            edits.append(ITUser.from_simplified_fields(**payload))

    await mo_model_client.edit(edits)
    await mo_model_client.upload(creates)


if __name__ == "__main__":
    cli()



from dataclasses import dataclass
from uuid import UUID

import uuids


@dataclass
class Class:
    titel: str
    facet: str
    scope: str
    bvn: str
    uuid: UUID


CLASSES = [
    Class(
        "Postadresse",
        "org_unit_address_type",
        "DAR",
        "AddressMailUnit",
        uuids.UNIT_POSTADDR,
    ),
    Class("LOS ID", "org_unit_address_type", "TEXT", "LOSID", uuids.UNIT_LOS),
    Class("CVR nummer", "org_unit_address_type", "TEXT", "CVRUnit", uuids.UNIT_CVR),
    Class("EAN nummer", "org_unit_address_type", "EAN", "EANUnit", uuids.UNIT_EAN),
    Class(
        "P-nummer",
        "org_unit_address_type",
        "PNUMBER",
        "PNumber",
        uuids.UNIT_PNR,
    ),
    Class("SE-nummer", "org_unit_address_type", "TEXT", "SENumber", uuids.UNIT_SENR),
    Class(
        "IntDebitor-Nr",
        "org_unit_address_type",
        "TEXT",
        "intdebit",
        uuids.UNIT_DEBITORNR,
    ),
    Class("WWW", "org_unit_address_type", "WWW", "UnitWeb", uuids.UNIT_WWW),
    Class(
        "Ekspeditionstid",
        "org_unit_address_type",
        "TEXT",
        "UnitHours",
        uuids.UNIT_HOURS,
    ),
    Class(
        "Telefontid",
        "org_unit_address_type",
        "TEXT",
        "UnitPhoneHours",
        uuids.UNIT_PHONEHOURS,
    ),
    Class(
        "Telefon",
        "org_unit_address_type",
        "PHONE",
        "UnitPhone",
        uuids.UNIT_PHONE,
    ),
    Class("Fax", "org_unit_address_type", "PHONE", "UnitFax", uuids.UNIT_FAX),
    Class("Email", "org_unit_address_type", "EMAIL", "UnitEmail", uuids.UNIT_EMAIL),
    Class(
        "Magkort",
        "org_unit_address_type",
        "TEXT",
        "UnitMagID",
        uuids.UNIT_MAG_ID,
    ),
    Class(
        "Alternativt navn",
        "org_unit_address_type",
        "TEXT",
        "UnitNameAlt",
        uuids.UNIT_NAME_ALT,
    ),
    Class(
        "Phone",
        "employee_address_type",
        "PHONE",
        "PhoneEmployee",
        uuids.PERSON_PHONE,
    ),
    Class(
        "Email",
        "employee_address_type",
        "EMAIL",
        "EmailEmployee",
        uuids.PERSON_EMAIL,
    ),
    Class(
        "Lokale",
        "employee_address_type",
        "TEXT",
        "RoomEmployee",
        uuids.PERSON_ROOM,
    ),
    Class("Primær", "primary_type", "100000", "primary", uuids.PRIMARY),
    Class("Ikke-primær", "primary_type", "0", "non-primary", uuids.NOT_PRIMARY),
    Class(
        "Linjeorganisation",
        "org_unit_hierarchy",
        "TEXT",
        "linjeorg",
        uuids.LINJE_ORG_HIERARCHY,
    ),
    Class(
        "Sikkerhedsorganisation",
        "org_unit_hierarchy",
        "TEXT",
        "sikkerhedsorg",
        uuids.SIKKERHEDS_ORG_HIERARCHY,
    ),
    Class(
        "Rolletype",
        "role_type",
        "TEXT",
        "role_type",
        UUID("964c31a2-6267-4388-bff5-42d6f3c5f708"),
    ),
    Class(
        "Tilknytningsrolle",
        "association_type",
        "TEXT",
        "association_type",
        UUID("ec534b86-3d9b-42d8-bff0-afc4f81719af"),
    ),
    Class(
        "Orlovstype",
        "leave_type",
        "TEXT",
        "leave_type",
        UUID("d2892fa6-bc56-4c14-bd24-74ae0c71fa3a"),
    ),
    Class(
        "Alternativ stillingsbetegnelse",
        "employee_address_type",
        "TEXT",
        "AltJobTitle",
        uuids.PERSON_JOB_TITLE_ALT,
    ),
]



from contextlib import nullcontext
from operator import itemgetter
from unittest import mock
from unittest import TestCase

from hypothesis import given
from hypothesis import settings
from hypothesis import strategies as st
from more_itertools import unique_everseen

from ..ad_reader import ADParameterReader
from .mocks import MockAD


AD_SAM_ACCOUNT_NAME = "SamAccountName"
AD_CPR_FIELD_NAME = "CprFieldName"


class _TestableADParameterReader(MockAD, ADParameterReader):
    def __init__(self, response, **overridden_settings):
        super().__init__()
        self._response = response
        # Populated by `ADParameterReader.uncached_read_user`
        self.results = {}
        # Read by `ADParameterReader.uncached_read_user` via `AD._get_setting`
        self.all_settings = {
            "global": {"servers": None},
            "primary": {
                "servers": None,
                "search_base": "",
                "properties": [AD_SAM_ACCOUNT_NAME, AD_CPR_FIELD_NAME],
                "cpr_field": AD_CPR_FIELD_NAME,
                "cpr_separator": None,
                "sam_filter": None,
                "caseless_samname": None,
            },
        }
        if overridden_settings:
            self.all_settings["primary"].update(overridden_settings)

    def get_from_ad(self, user=None, cpr=None, server=None):
        return self._response


class TestADParameterReader(TestCase):
    """Test `ADParameterReader`"""

    @settings(max_examples=1000, deadline=None)
    @given(
        # Build simulated AD response
        st.lists(
            st.fixed_dictionaries(
                {
                    AD_SAM_ACCOUNT_NAME: st.text(min_size=1),
                    AD_CPR_FIELD_NAME: st.text() | st.none(),
                }
            )
        ),
        # Build simulated settings
        st.fixed_dictionaries(
            {
                "cpr_separator": st.text(max_size=1),
                "caseless_samname": st.booleans(),
                "sam_filter": st.text(),
            }
        ),
        # Decide whether `first_included` returns an AD user or an empty dict
        st.booleans(),
    )
    def test_uncached_read_user(self, response, settings, first_included_is_empty):
        reader = _TestableADParameterReader(response, **settings)
        ria = []  # accumulates all AD users found
        user = "foobar"

        with self._mock_first_included(first_included_is_empty):
            reader.uncached_read_user(user=user, ria=ria)

        # Test contents of the `ria` accumulator variable
        if first_included_is_empty:
            # `ria` is expected to be empty if `first_included` returns an
            # empty dict.
            expected_ad_users = []
        else:
            # `ria` must contain an AD user for each non-blank SamAccountName
            # *or* CPR. Each AD user must only show up *once* in `ria`.
            expected_ad_users = list(
                unique_everseen(
                    filter(
                        lambda ad_user: (
                            ad_user.get(AD_SAM_ACCOUNT_NAME)
                            or ad_user.get(AD_CPR_FIELD_NAME)
                        ),
                        response,
                    ),
                    key=itemgetter(AD_CPR_FIELD_NAME),
                )
            )
        self.assertListEqual(
            ria,
            expected_ad_users,
            "\nria:\nactual:   %r\nexpected: %r" % (ria, expected_ad_users),
        )

        # TODO: Test contents of `reader.results`
        self.assertTrue(
            len(reader.results) > 0
            if (response and not first_included_is_empty)
            else len(reader.results) == 0
        )

    def _mock_first_included(self, first_included_is_empty):
        if first_included_is_empty:
            path = "integrations.ad_integration.ad_reader.first_included"
            return mock.patch(path, return_value={})
        else:
            return nullcontext()



import uuid
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Type

import pytest
from os2mo_helpers.mora_helpers import MoraHelper

from ..ad_reader import ADParameterReader
from ..import_ad_group_into_mo import ADMOImporter
from .mocks import MO_AD_IT_SYSTEM_UUID
from .mocks import MO_ROOT_ORG_UNIT_UUID
from .mocks import MockADParameterReader
from .mocks import MockMoraHelper


class _MockMoraHelperNoEmployeeITUsers(MockMoraHelper):
    """Mock a `MoraHelper` which has a MO employee, but that employee has no IT users
    nor any engagements.
    """

    def get_e_itsystems(self, e_uuid, it_system_uuid=None):
        return []

    def read_user_engagement(self, *args, **kwargs):
        return []


class _MockMoraHelperNoEmployee(_MockMoraHelperNoEmployeeITUsers):
    """Mock a `MoraHelper` which hasn't got the MO employee that we try to read."""

    def read_user(self, user_cpr=None, **kwargs):
        return []


class _MockADParameterReaderResults(MockADParameterReader):
    """Mock an AD containing a single user, having just enough information to exercise
    `ADMOImporter`.
    """

    @property
    def results(self) -> Dict[str, Dict]:
        return {
            "bob": {
                # `DistinguishedName` value must be a substring of
                # `integrations.ad.import_ou.search_string` in `_TestableADMOImporter`.
                "DistinguishedName": "CN=Bob,OU=Konsulenter,DC=Magenta,DC=dk",
                "SamAccountName": "bob",
                "Cpr": "bob-cpr",
                "ObjectGUID": str(uuid.uuid4()),
                "GivenName": "Bob",
                "Surname": "Bobson",
            }
        }


class _TestableADMOImporter(ADMOImporter):
    def __init__(self, _mora_helper_class: Type[MockMoraHelper]):
        self._mora_helper_class = _mora_helper_class
        super().__init__()

    def _get_settings(self) -> Dict[str, Any]:
        return {
            "integrations.ad": [{"cpr_field": "Cpr"}],
            "integrations.ad.import_ou.mo_unit_uuid": str(MO_ROOT_ORG_UNIT_UUID),
            "integrations.ad.import_ou.search_string": "Konsulent",
        }

    def _get_mora_helper(self) -> MoraHelper:
        return self._mora_helper_class(None, mo_uuid=uuid.uuid4())

    def _get_ad_it_system(self) -> uuid.UUID:
        return MO_AD_IT_SYSTEM_UUID

    def _get_ad_reader(self) -> ADParameterReader:
        return _MockADParameterReaderResults()  # type: ignore

    def _ensure_class_in_lora(
        self, facet_bvn: str, class_value: str
    ) -> Tuple[uuid.UUID, bool]:
        return uuid.uuid4(), True


class TestADMOImporter:
    @pytest.mark.parametrize(
        "mora_helper_class,expected_mo_post_calls",
        [
            # Case 1: MO employee does not exist - create employee, IT user and
            # engagement.
            (
                _MockMoraHelperNoEmployee,
                [
                    {
                        "url": "e/create",
                        "payload": {
                            "cpr_no": "bobcpr",
                            "givenname": "Bob",
                            "surname": "Bobson",
                        },
                        "force": True,
                    },
                    {
                        "url": "details/create",
                        "payload": {"type": "it", "user_key": "bob"},
                        "force": True,
                    },
                    {
                        "url": "details/create",
                        "payload": {"type": "engagement", "user_key": "bob"},
                        "force": True,
                    },
                ],
            ),
            # Case 2: MO employee exists, but IT user and engagement do not - create
            # IT user and engagement.
            (
                _MockMoraHelperNoEmployeeITUsers,
                [
                    {
                        "url": "details/create",
                        "payload": {"type": "it", "user_key": "bob"},
                        "force": True,
                    },
                    {
                        "url": "details/create",
                        "payload": {"type": "engagement", "user_key": "bob"},
                        "force": True,
                    },
                ],
            ),
        ],
    )
    def test_create_or_update_users_in_mo(
        self,
        mora_helper_class: MockMoraHelper,
        expected_mo_post_calls: List[dict],
    ):
        # Regression test for #51052
        instance = _TestableADMOImporter(_mora_helper_class=mora_helper_class)
        instance.create_or_update_users_in_mo()

        assert len(expected_mo_post_calls) == len(instance.helper._mo_post_calls)
        paired_calls = zip(expected_mo_post_calls, instance.helper._mo_post_calls)
        for expected_call, actual_call in paired_calls:
            # Assert that `expected_call` contains a subset of `actual_call`
            # See: https://github.com/pytest-dev/pytest/issues/2376#issuecomment-852366588
            assert {**actual_call, **expected_call} == expected_call



import datetime
import hashlib
import logging
import pickle
import uuid
from enum import Enum
from functools import lru_cache
from functools import wraps
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import OrderedDict
from typing import Union

import requests
import xmltodict
from ra_utils.load_settings import load_settings

from .config import CommonSettings
from .config import get_common_settings


logger = logging.getLogger("sdCommon")


@lru_cache(maxsize=None)
def sd_lookup_settings():
    settings = load_settings()

    institution_identifier = settings["integrations.SD_Lon.institution_identifier"]
    if not institution_identifier:
        raise ValueError("Missing setting, institution_identifier")

    sd_user = settings["integrations.SD_Lon.sd_user"]
    if not sd_user:
        raise ValueError("Missing setting, sd_user")

    sd_password = settings["integrations.SD_Lon.sd_password"]
    if not sd_password:
        raise ValueError("Missing setting, sd_password")

    return institution_identifier, sd_user, sd_password


def _sd_lookup_cache(func):
    def create_hex_digest(full_url, payload):
        """Create a reproducible hex digest from url and payloads."""
        hasher = hashlib.sha256()

        for key, value in sorted(payload.items()):
            hasher.update((str(key) + str(value)).encode())
        hasher.update(full_url.encode())

        return hasher.hexdigest()

    def write_response(cache_file, response):
        """Write response to disk."""
        with open(str(cache_file), "wb") as f:
            pickle.dump(response, f, pickle.HIGHEST_PROTOCOL)

    def read_response(cache_file):
        """Read response from disk."""
        with open(str(cache_file), "rb") as f:
            response = pickle.load(f)
        return response

    @wraps(func)
    def wrapper(full_url, payload, auth, use_cache=True):
        # Short-circuit as noop, if no caching is requested
        if use_cache is False:
            return func(full_url, payload, auth)

        # We need a cache dir to exist before we can proceed
        cache_dir = Path(get_common_settings().cache_folder_path)
        if not cache_dir.is_dir():
            raise Exception("Folder for temporary files does not exist")

        # Create digest and find filename
        lookup_id = create_hex_digest(full_url, payload)
        cache_file = cache_dir / ("sd_" + lookup_id + ".p")

        # If cache file was found, use it
        if cache_file.is_file():
            response = read_response(cache_file)
            logger.info("This SD lookup was found in cache: {}".format(lookup_id))
            logger.debug(full_url + " - read from cache")
        else:  # No cache
            response = func(full_url, payload, auth)
            write_response(cache_file, response)
            logger.debug(full_url + " - requested from SD")
        return response

    return wrapper


@_sd_lookup_cache
def _sd_request(full_url, payload, auth):
    """Fire the actual request against SD.

    Annotation only calls this if we did not hit the cache.
    """
    return requests.get(
        full_url,
        params=payload,
        auth=auth,
    )


def sd_lookup(
    url: str,
    settings: Optional[CommonSettings] = None,
    params: Optional[Dict[str, Any]] = None,
    use_cache: bool = True,
) -> OrderedDict:
    """Fire a requests against SD.

    Utilizes _sd_request to fire the actual request, which in turn utilize
    _sd_lookup_cache for caching.
    """
    logger.info("Retrieve: {}".format(url))
    logger.debug("Params: {}".format(params))

    params = params or dict()

    BASE_URL = "https://service.sd.dk/sdws/"
    full_url = BASE_URL + url

    # Use settings if provided as an argument to this function
    # Currently, we only have Pydantic settings models for sd_importer.py
    # and sd_changed_at.py, so this logic is required for now, since the
    # sd_lookup function is used elsewhere too
    if settings is not None:
        institution_identifier = settings.sd_institution_identifier
        sd_user = settings.sd_user
        sd_password = settings.sd_password.get_secret_value()
    else:
        institution_identifier, sd_user, sd_password = sd_lookup_settings()

    payload = {
        "InstitutionIdentifier": institution_identifier,
    }
    payload.update(params)
    auth = (sd_user, sd_password)
    response = _sd_request(full_url, payload, auth, use_cache=use_cache)
    logger.debug("Response: {}".format(response.text))

    dict_response = xmltodict.parse(response.text)

    if url in dict_response:
        xml_response = dict_response[url]
    else:
        msg = "SD api error, envelope: {}, response: {}"
        logger.error(msg.format(dict_response["Envelope"], response.text))
        raise Exception(msg.format(dict_response["Envelope"], response.text))
    logger.debug("Done with {}".format(url))
    return xml_response


def calc_employment_id(employment):
    employment_id = employment["EmploymentIdentifier"]
    try:
        employment_number = int(employment_id)
    except ValueError:  # Job id is not a number?
        employment_number = 999999

    employment_id = {"id": employment_id, "value": employment_number}
    return employment_id


def mora_assert(response):
    """Check response is as expected."""
    assert response.status_code in (200, 201, 400, 404), response.status_code
    if response.status_code == 400:
        # Check actual response
        assert (
            response.text.find("not give raise to a new registration") > 0
        ), response.text
        logger.debug("Request had no effect")
    return None


def generate_uuid(value, org_id_prefix, org_name=None):
    """
    Code almost identical to this also lives in the Opus importer.
    """
    # TODO: Refactor to avoid duplication
    if org_id_prefix:
        base_hash = hashlib.md5(org_id_prefix.encode())
    else:
        base_hash = hashlib.md5(org_name.encode())

    base_digest = base_hash.hexdigest()
    base_uuid = uuid.UUID(base_digest)

    combined_value = (str(base_uuid) + str(value)).encode()
    value_hash = hashlib.md5(combined_value)
    value_digest = value_hash.hexdigest()
    value_uuid = str(uuid.UUID(value_digest))
    return value_uuid


def primary_types(helper):
    """
    Read the engagement types from MO and match them up against the four
    known types in the SD->MO import.

    Args:
        helper: An instance of mora-helpers.

    Returns:
        A dict matching up the engagement types with LoRa class uuids (
        i.e. UUIDs of facets).

    Example:
        An example return value:
        ```python
        {
            "primary": "697c8838-ba0f-4e74-90f8-4e7c31d4e7e7",
            "non_primary": "b9543f90-9511-494b-bbf5-f15678502c2d",
            "no_salary": "88589e84-5736-4f8c-9c0c-2e29046d7471",
            "fixed_primary": "c95a1999-9f95-4458-a218-e9c96e7ad3db",
        }
        ```
    """

    # These constants are global in all SD municipalities (because they are created
    # by the SD->MO importer.
    PRIMARY = "Ansat"
    NO_SALARY = "status0"
    NON_PRIMARY = "non-primary"
    FIXED_PRIMARY = "explicitly-primary"

    logger.info("Read primary types")
    primary = None
    no_salary = None
    non_primary = None
    fixed_primary = None

    primary_types = helper.read_classes_in_facet("primary_type")
    for primary_type in primary_types[0]:
        if primary_type["user_key"] == PRIMARY:
            primary = primary_type["uuid"]
        if primary_type["user_key"] == NON_PRIMARY:
            non_primary = primary_type["uuid"]
        if primary_type["user_key"] == NO_SALARY:
            no_salary = primary_type["uuid"]
        if primary_type["user_key"] == FIXED_PRIMARY:
            fixed_primary = primary_type["uuid"]

    type_uuids = {
        "primary": primary,
        "non_primary": non_primary,
        "no_salary": no_salary,
        "fixed_primary": fixed_primary,
    }
    if None in type_uuids.values():
        raise Exception("Missing primary types: {}".format(type_uuids))
    return type_uuids


class EmploymentStatus(Enum):
    """Corresponds to EmploymentStatusCode from SD.

    Employees usually start in AnsatUdenLoen, and then change to AnsatMedLoen.
    This will usually happen once they actually have their first day at work.

    From AnsatMedLoen they can somewhat freely transfer to the other statusses.
    This includes transfering back to AnsatMedLoen from any other status.

    Note for instance, that it is entirely possible to be Ophoert and then get
    hired back, and thus go from Ophoert to AnsatMedLoen.

    There is only one terminal state, namely Slettet, wherefrom noone will
    return. This state is invoked from status 7-8-9 after a few years.

    Status Doed will probably only migrate to status slettet, but there are no
    guarantees given.
    """

    # This status most likely represent not yet being at work
    AnsatUdenLoen = "0"

    # These statusses represent being at work
    AnsatMedLoen = "1"
    Orlov = "3"

    # These statusses represent being let go
    Migreret = "7"
    Ophoert = "8"
    Doed = "9"

    # This status is the special terminal state
    Slettet = "S"

    @staticmethod
    def employeed() -> List["EmploymentStatus"]:
        return [
            EmploymentStatus.AnsatUdenLoen,
            EmploymentStatus.AnsatMedLoen,
            EmploymentStatus.Orlov,
        ]

    @staticmethod
    def let_go() -> List["EmploymentStatus"]:
        return [
            EmploymentStatus.Migreret,
            EmploymentStatus.Ophoert,
            EmploymentStatus.Doed,
        ]

    @staticmethod
    def on_payroll() -> List["EmploymentStatus"]:
        return [EmploymentStatus.AnsatMedLoen, EmploymentStatus.Orlov]


def skip_fictional_users(entity):
    cpr = entity["PersonCivilRegistrationIdentifier"]
    if cpr[-4:] == "0000":
        logger.warning("Skipping fictional user: {}".format(cpr))
        return False
    return True


def ensure_list(element):
    if not isinstance(element, list):
        return [element]
    return element


# We will get to the Pydantic models later...
def read_employment_at(
    effective_date: datetime.date,
    settings: CommonSettings,
    employment_id: Optional[str] = None,
    status_active_indicator: bool = True,
    status_passive_indicator: bool = True,
) -> Union[OrderedDict, List[OrderedDict], None]:
    url = "GetEmployment20111201"
    params = {
        "EffectiveDate": effective_date.strftime("%d.%m.%Y"),
        "StatusActiveIndicator": str(status_active_indicator).lower(),
        "StatusPassiveIndicator": str(status_passive_indicator).lower(),
        "DepartmentIndicator": "true",
        "EmploymentStatusIndicator": "true",
        "ProfessionIndicator": "true",
        "WorkingTimeIndicator": "true",
        "UUIDIndicator": "true",
        "SalaryAgreementIndicator": "false",
        "SalaryCodeGroupIndicator": "false",
    }

    if employment_id:
        params.update({"EmploymentIdentifier": employment_id})

    response = sd_lookup(url, settings=settings, params=params)
    return response.get("Person")



import csv
import datetime
import hashlib
import json
import logging
import pathlib
import pickle
import uuid
from logging.handlers import RotatingFileHandler

import requests
from anytree import Node
from chardet.universaldetector import UniversalDetector

INFO_LEVEL = 20
LOG_FILE = 'udvalg.log'
BASE_URL = 'http://localhost:5000/service/'
CACHE = {}
SESSION = None 

logger = logging.getLogger()
log_format = logging.Formatter(
    '%(asctime)s | %(name)s |  %(levelname)s: %(message)s'
)
logger.setLevel(logging.DEBUG)
activity_log_handler = RotatingFileHandler(
    filename=LOG_FILE,
    maxBytes=1000000
)
activity_log_handler.setFormatter(log_format)
activity_log_handler.setLevel(INFO_LEVEL)
logger.addHandler(activity_log_handler)


def _find_class(find_facet, find_class):
    if find_class in CACHE:
        return CACHE[find_class]
    uuid = None
    url = BASE_URL + 'o/{}/f/{}'
    response = SESSION.get(url.format(ROOT, find_facet)).json()
    for actual_class in response['data']['items']:
        if actual_class['user_key'] == find_class:
            uuid = actual_class['uuid']
            CACHE[find_class] = uuid
    if uuid:
        return uuid
    raise Exception("Der blev ikke fundet et uuid for denne" )


def _mo_lookup(uuid, details=''):
    if not details:
        url = BASE_URL + 'e/{}'
    else:
        url = BASE_URL + 'e/{}/details/' + details
    response = SESSION.get(url.format(uuid))
    return(response.json())


def _find_org():
    url = BASE_URL + 'o'
    response = SESSION.get(url).json()
    assert(len(response) == 1)
    uuid = response[0]['uuid']
    return uuid

def _load_csv(file_name):
    rows = []
    detector = UniversalDetector()
    with open(file_name, 'rb') as csvfile:
        for row in csvfile:
            detector.feed(row)
            if detector.done:
                break
    detector.close()
    encoding = detector.result['encoding']

    with open(file_name, encoding=encoding) as csvfile:
        reader = csv.DictReader(csvfile, delimiter=';')
        for row in reader:
            rows.append(row)
    return rows


def generate_uuid(value, org_name):
    """
    Code almost identical to this also lives in the Opus importer.
    """
    base_hash = hashlib.md5(org_name.encode())

    base_digest = base_hash.hexdigest()
    base_uuid = uuid.UUID(base_digest)

    combined_value = (str(base_uuid) + str(value)).encode()
    value_hash = hashlib.md5(combined_value)
    value_digest = value_hash.hexdigest()
    value_uuid = str(uuid.UUID(value_digest))
    return value_uuid


def _create_mo_ou(name, parent, org_type, bvn):
    uuid = generate_uuid(bvn, ROOT)
    ou_type = _find_class(find_facet='org_unit_type', find_class=org_type)
    ou_level = _find_class(find_facet='org_unit_level', find_class='MED-enhed')
    if parent == 'root':
        parent = ROOT
    payload = {
        'uuid': uuid,
        'user_key': str(bvn),
        'name': '{} {}'.format(org_type, name),
        'org_unit_type': {'uuid': ou_type},
        'org_unit_level':{'uuid': ou_level},
        'parent': {'uuid': parent},
        'validity': {'from': '1930-01-01',
                     'to':  None}
    }

    url = BASE_URL + 'ou/create'
    params = {'force': 1}
    response = SESSION.post(url, json=payload, params=params)
    uuid = response.json()
    return uuid


def _create_mo_association(user, org_unit, association_type, from_string):
    response = _mo_lookup(user, details='engagement')
    if response:
        job_function = response[0]['job_function']['uuid']
        payload = [
            {
                'type': 'association',
                'org_unit': {'uuid': org_unit},
                'person': {'uuid': user},
                'association_type': {'uuid': association_type},
                'job_function': {'uuid': job_function},
                'validity': {
                    'from': from_string,
                    'to': None
                }
            }
        ]
        url = BASE_URL + 'details/create'
        params = {'force': 1}

        response = SESSION.post(url, json=payload, params=params)
        response.raise_for_status()
        uuid = response.json()
        return uuid
    else:
        logger.warning('User {} has no engagements'.format(user))
        return None


def _create_mo_role(user, org_unit, role_type, from_string):
    response = _mo_lookup(user, details='engagement')
    if response:
        job_function = response[0]['job_function']['uuid']
        payload = [
            {
                'type': 'role',
                'org_unit': {'uuid': org_unit},
                'person': {'uuid': user},
                'role_type': {'uuid': role_type},
                'job_function': {'uuid': job_function},
                'validity': {
                    'from': from_string,
                    'to': None
                }
            }
        ]
        url = BASE_URL + 'details/create'
        params = {'force': 1}
        response = SESSION.post(url, json=payload, params=params)
        response.raise_for_status()
        uuid = response.json()
        return uuid
    else:
        logger.warning('User {} has no engagements'.format(user))
        return None


def create_udvalg(nodes, file_name):
    rows = _load_csv(file_name)
    for row in rows:
        if ('Formand' in row) and (row['Formand'] == '1'):
            association_type = _find_class('association_type', 'MED-Formand')
        elif ('Næstformand' in row) and (row['Næstformand'] == '1'):
            association_type = _find_class('association_type', 'MED-Næstformand')
        elif ('suppleant' in row) and (row['suppleant'] == '1'):
            association_type = _find_class('association_type', 'MED-Medlem-Suppleant')    
        else:
            association_type = _find_class('association_type', 'MED-Medlem')

        if ('TR' in row) and (row['TR'] == '1'):
            role_type = _find_class('role_type', 'MED-Tillidsrepræsentant')
        else:
            role_type = None

        org_id = int(row['Id'])
        uuid = row['BrugerUUID']
        try:
            from_string = datetime.datetime.strftime(
                datetime.datetime.strptime(row['StartDato'], '%d-%b-%y'),
                '%Y-%m-%d'
            )
        except ValueError:
            from_string = '1930-01-01'
        except TypeError:
            from_string = '1930-01-01'

        if uuid:
            nodes[uuid] = Node(row['Fornavn'] + ' ' + row['Efternavn'],
                               uuid=uuid,
                               org_type=row['OrgType'],
                               parent=nodes[org_id])
            _create_mo_association(uuid,
                                   nodes[org_id].uuid,
                                   association_type,
                                   from_string)
            if role_type:
                _create_mo_role(uuid, nodes[org_id].uuid, role_type, from_string)

        else:
            logger.warning(
                'Error: {} {}, UUID: {}'.format(row['Fornavn'],
                                               row['Efternavn'],
                                               row['BrugerUUID'])
            )
    return nodes


def create_tree(file_name):
    nodes = {}
    rows = _load_csv(file_name)
    while rows:
        new = {}
        remaining_nodes = []
        for row in rows:
            org_type = row['OrgType'].strip()
            id_nr = int(row['Id'])
            parent = int(row['ParentID']) if row['ParentID'] else None
            if parent is None:
                uuid = _create_mo_ou(row['OrgEnhed'], parent='root',
                                     org_type=org_type, bvn=id_nr)
                new[id_nr] = Node(row['OrgEnhed'],
                                  uuid=uuid, org_type=org_type)
            elif parent in nodes:
                uuid = _create_mo_ou(row['OrgEnhed'],
                                     parent=nodes[parent].uuid,
                                     org_type=org_type, bvn=id_nr)
                new[id_nr] = Node(row['OrgEnhed'],
                                  uuid=uuid, org_type=org_type,
                                  parent=nodes[parent])
            else:
                remaining_nodes.append(row)
        rows = remaining_nodes
        nodes.update(new)
    return nodes


if __name__ == '__main__':
    logger.info('Program started')

    settingsfile = pathlib.Path("settings") / "settings.json"
    settings = json.loads(settingsfile.read_text())


    SESSION = requests.Session()
    if settings.get("crontab.SAML_TOKEN", None) is not None:
        SESSION.headers["SESSION"] = settings["crontab.SAML_TOKEN"]


    ROOT = _find_org()

    orgtyper_file = '/opt/settings/MED-org/MED-Organisation.csv'
    med_medlemmer_file = '/opt/settings/MED-org/MED-Medlemmer.csv'
    
    if True:
        nodes = create_tree(orgtyper_file)
        with open('nodes.p', 'wb') as f:
            pickle.dump(nodes, f, pickle.HIGHEST_PROTOCOL)

    with open('nodes.p', 'rb') as f:
        nodes = pickle.load(f)


    logger.info('Create MED')
    nodes = create_udvalg(nodes, med_medlemmer_file)

    # root = min(nodes.keys())
    # from anytree import RenderTree
    # print(RenderTree(nodes[root]))
    logger.info('Program completed.')



import csv
from operator import itemgetter


def employee_mapper(filename:str) -> dict:
    """
    >>> import tempfile
    >>> with tempfile.NamedTemporaryFile('w', encoding='utf-8') as tmpfile:
    ...      _ = tmpfile.write('cpr;mo_uuid;ad_guid;sam_account_name\\n')
    ...      _ = tmpfile.write('0101701111;1034f987-555a-4183-bad1-2c5846117cee;;\\n')
    ...      _ = tmpfile.write('1103701234;a9077886-db39-400a-b3c4-8b73933790d2;;\\n')
    ...      tmpfile.flush()
    ...      expected = {
    ...         '0101701111': '1034f987-555a-4183-bad1-2c5846117cee',
    ...         '1103701234': 'a9077886-db39-400a-b3c4-8b73933790d2'
    ...      }
    ...      employee_mapper(tmpfile.name) == expected
    True
    """
    with open(filename) as csvfile:
        reader = csv.DictReader(csvfile, delimiter=';')
        return dict(map(itemgetter('cpr', 'mo_uuid'), reader))  # type: ignore



from typing import List, Optional, Tuple

import asyncio
from collections import ChainMap
from functools import partial
from operator import itemgetter

from aiohttp import ClientSession, TCPConnector
from more_itertools import chunked, unzip

from ra_utils.async_to_sync import async_to_sync


def ensure_session(func):
    async def _decorator(*args, **kwargs):
        if "client" in kwargs:
            return await func(*args, **kwargs)
        else:
            # DAWA only accepts:
            # * 30 requests per second per IP
            # * 10 concurrent connections per IP
            # Thus we limit our connections to 10 here.
            connector = TCPConnector(limit=10)
            async with ClientSession(connector=connector) as client:
                return await func(*args, **kwargs, client=client)

    return _decorator


@ensure_session
async def dar_fetch_non_chunked(uuids, addrtype, client=None):
    """Lookup uuids in DAR (without chunking).

    Args:
        uuids: List of DAR UUIDs to lookup.
        addr_type: The address type to lookup.
        client (optional): aiohttp.ClientSession to use for connecting.

    Returns:
        (dict, set):
            dict: Map from UUID to DAR reply.
            set: Set of UUIDs of entries which were not found.
    """
    url = "https://dawa.aws.dk/" + addrtype
    params = {"id": "|".join(uuids), "struktur": "mini"}

    async with client.get(url, params=params) as response:
        response.raise_for_status()
        body = await response.json()

        result = {addr['id']: addr for addr in body}

        found_uuids = set(map(itemgetter("id"), body))
        missing = set(uuids) - found_uuids

        return result, missing


@ensure_session
async def dar_fetch_chunked(uuids, addrtype, chunk_size, client=None):
    """Lookup uuids in DAR (chunked).

    Args:
        uuids: List of DAR UUIDs.
        addr_type: The address type to lookup.
        chunk_size: Number of UUIDs per block, sent to DAR.
        client (optional): aiohttp.ClientSession to use for connecting.

    Returns:
        (dict, set):
            dict: Map from UUID to DAR reply.
            set: Set of UUIDs of entries which were not found.
    """

    def create_task(uuid_chunk):
        return asyncio.ensure_future(
            dar_fetch_non_chunked(uuid_chunk, addrtype=addrtype, client=client)
        )

    # Chunk our UUIDs into blocks of chunk_size
    uuid_chunks = chunked(uuids, chunk_size)
    # Convert chunks into a list of asyncio.tasks
    tasks = list(map(create_task, uuid_chunks))
    # Here 'result' is a list of tuples (dict, set) => (result, missing)
    result = await asyncio.gather(*tasks)
    # First we unzip 'result' to get a list of results and a list of missing
    result_dicts, missing_sets = unzip(result)
    # Then we union the dicts and sets before returning
    combined_result = dict(ChainMap(*result_dicts))
    combined_missing = set.union(*missing_sets)
    return combined_result, combined_missing


@ensure_session
async def dar_fetch(uuids, addrtype="adresser", chunk_size=150, client=None):
    """Lookup uuids in DAR (chunked if required).

    Args:
        uuids: List of DAR UUIDs.
        addr_type: The address type to lookup.
        chunk_size: Number of UUIDs per block, sent to DAR.
        client (optional): aiohttp.ClientSession to use for connecting.

    Returns:
        (dict, set):
            dict: Map from UUID to DAR reply.
            set: Set of UUIDs of entries which were not found.
    """
    num_uuids = len(uuids)
    if num_uuids == 0:
        return dict(), set()
    elif num_uuids <= 150:
        return await dar_fetch_non_chunked(uuids, addrtype, client=client)
    else:
        return await dar_fetch_chunked(uuids, addrtype, chunk_size, client=client)


@async_to_sync
async def sync_dar_fetch(uuids, addrtype="adresser", chunk_size=150):
    """Syncronized version of dar_fetch."""
    return await dar_fetch(uuids, addrtype, chunk_size)


@ensure_session
async def dar_datavask(address: str, client: ClientSession) -> Tuple[str, Optional[dict]]:
    """
    Perform search in DAR using the 'datavask' API
    """
    url = "https://api.dataforsyningen.dk/datavask/adresser"
    params = {"struktur": "mini", "betegnelse": address}
    async with client.get(url, params=params) as response:
        response.raise_for_status()
        result = await response.json()
        # A and B are safe matches, and will only have one result in the array
        # https://dawadocs.dataforsyningen.dk/dok/api/adresse#datavask
        if result['kategori'] in ['A', 'B']:
            return address, result['resultater'][0]['adresse']['id']

    return address, None


@ensure_session
async def dar_datavask_multiple(addresses: List[str], client: ClientSession) -> List[Tuple[str, Optional[dict]]]:
    """Perform search in DAR on multiple address strings"""
    tasks = map(partial(dar_datavask, client=client), addresses)
    results = await asyncio.gather(*tasks)
    return results



import logging

import click
import requests
from os2mo_helpers.mora_helpers import MoraHelper

from .config import CommonSettings
from .models import JobFunction
from .sd_common import mora_assert
from .sd_common import sd_lookup
from .sd_payloads import edit_klasse_title


LOG_LEVEL = logging.DEBUG
LOG_FILE = "sync_job_id.log"

logger = logging.getLogger("sdSyncJobId")


def setup_logging():
    for name in logging.root.manager.loggerDict:
        if name in ("sdSyncJobId", "sdCommon", "mora-helper"):
            logging.getLogger(name).setLevel(LOG_LEVEL)
        else:
            logging.getLogger(name).setLevel(logging.ERROR)

    logging.basicConfig(
        format="%(levelname)s %(asctime)s %(name)s %(message)s",
        level=LOG_LEVEL,
        filename=LOG_FILE,
    )


class JobIdSync:
    def __init__(self, settings: CommonSettings):
        logger.info("Start sync")
        self.settings = settings

        sd_job_function = self.settings.sd_job_function
        if sd_job_function == JobFunction.job_position_identifier:
            logger.info("Read settings. Update job_functions and engagment types")
            self.update_job_functions = True
        else:
            logger.info("Read settings. Do not update job_functions")
            self.update_job_functions = False

        self._read_classes()

    def _read_classes(self):
        """Read engagement_types and job_function types from MO."""
        mora_base = self.settings.mora_base
        helper = MoraHelper(hostname=mora_base, use_cache=False)

        self.engagement_types = helper.read_classes_in_facet("engagement_type")
        if self.update_job_functions:
            self.job_function_types = helper.read_classes_in_facet(
                "engagement_job_function"
            )

    def _find_engagement_type(self, job_pos_id):
        """
        Find a Klasse in facet engagement_type corresponding to job_pos_id in LoRa.
        The search is performed both by direct search as well as with prefixed
        string engagement_type (which is used in the municipalities who also use
        the job_pos_ids as job_functions).
        """
        logger.info("Search MO for engagment_type {}".format(job_pos_id))
        found_type = None
        user_keys = [str(job_pos_id), "engagement_type" + str(job_pos_id)]
        for engagement_type in self.engagement_types[0]:
            if engagement_type["user_key"] in user_keys:
                found_type = engagement_type
        logger.info("Found {}".format(found_type))
        return found_type

    def _find_job_function_type(self, job_pos_id):
        """
        Find the Klasse corresponding to job_pos_id in LoRa.
        """
        found_type = None
        if not self.update_job_functions:
            logger.info("Job functons not enabled in settings")
            return None

        logger.info("Search MO for job_function_type {}".format(job_pos_id))
        # Currently we do not use a prefix anywhere, list has only one element
        user_keys = [str(job_pos_id)]
        for job_function_type in self.job_function_types[0]:
            if job_function_type["user_key"] in user_keys:
                found_type = job_function_type
        logger.info("Found {}".format(found_type))
        return found_type

    def _edit_klasse_title(self, uuid, title):
        """
        Change the title of an existing LoRa engagement type.
        """
        logger.info("Edit {} to {}".format(uuid, title))
        payload = edit_klasse_title(title)
        response = requests.patch(
            url=self.settings.mox_base + "/klassifikation/klasse/" + uuid,
            json=payload,
        )
        logger.info("Lora response: {}".format(response.status_code))
        mora_assert(response)
        return response

    def _get_job_pos_id_from_sd(self, job_pos_id):
        """
        Return the textual value of a Job Position Identifier from SD.
        """
        logger.info("Search SD for {}".format(job_pos_id))
        params = {
            "JobPositionIdentifier": job_pos_id,
        }
        try:
            job_pos_response = sd_lookup("GetProfession20080201", params)
        except Exception:  # TODO: Be specific here
            logger.info("This job_position could not be found in SD")
            return None
        job_pos = None
        while "Profession" in job_pos_response:
            job_pos = job_pos_response["Profession"]["JobPositionName"]
            job_pos_response = job_pos_response["Profession"]
        logger.info("Found {}".format(job_pos_id))
        return job_pos

    def _sync_engagement_type_from_sd(self, job_pos_id, sd_job_pos_text):

        mo_eng_type = self._find_engagement_type(job_pos_id)
        if mo_eng_type is None:
            logger.info("Engagement type {} not found i MO".format(job_pos_id))
            return False

        self._edit_klasse_title(mo_eng_type["uuid"], sd_job_pos_text)
        logger.info("Updated engagement type: {}".format(job_pos_id))
        return True

    def _sync_job_function_from_sd(self, job_pos_id, sd_job_pos_text):

        mo_job_function_type = self._find_job_function_type(job_pos_id)
        if mo_job_function_type is None:
            logger.info("job function type {} not found i MO".format(job_pos_id))
            return False

        self._edit_klasse_title(mo_job_function_type["uuid"], sd_job_pos_text)
        logger.info("Updated job function type type: {}".format(job_pos_id))
        return True

    def sync_from_sd(self, job_pos_id, refresh=False):
        """
        Sync the titel of LoRa engagement type to the value current
        registred at SD.
        """
        # If asked to refresh, reread the classes from MO. This may be necessary if
        # new classes have been added since the creation of this JobIdSync object.
        if refresh:
            self._read_classes()

        logger.info("Sync {} to value found in SD".format(job_pos_id))
        return_status = [None, None]

        sd_job_pos_text = self._get_job_pos_id_from_sd(job_pos_id)
        if sd_job_pos_text is None:
            logger.info("Job position {} not found i SD".format(job_pos_id))
            return return_status

        return_status[0] = self._sync_engagement_type_from_sd(
            job_pos_id, sd_job_pos_text
        )
        # Only run this part, if we are actually using
        if self.update_job_functions:
            return_status[1] = self._sync_job_function_from_sd(
                job_pos_id, sd_job_pos_text
            )

        logger.info("Return status: {}".format(return_status))
        return return_status

    def sync_all_from_sd(self):
        logger.info("Sync all classes")
        for eng_type in self.engagement_types[0]:
            user_key = eng_type["user_key"]
            if user_key.startswith("engagement_type"):
                user_key = user_key[15:]
            print("Sync from SD: {}".format(user_key))
            self.sync_from_sd(user_key)

        if self.update_job_functions:
            for job_function in self.job_function_types[0]:
                user_key = job_function["user_key"]
                print("Sync from SD: {}".format(user_key))
                self.sync_from_sd(user_key)
        logger.info("Full sync completed")

    def sync_manually(self, job_pos_id, title):
        """
        Manually update the titel of an engagement type.
        """
        logger.info("Sync {} to {}".format(job_pos_id, title))
        return_status = [None, None]

        mo_type = self._find_engagement_type(job_pos_id)
        if mo_type is None:
            return_status[0] = False
            logger.info("Job position not found i MO")
        else:
            return_status[0] = True
            self._edit_klasse_titel(mo_type["uuid"], title)

        if self.update_job_functions:
            mo_job_function_type = self._find_job_function_type(job_pos_id)
            if mo_job_function_type is None:
                return_status[1] = False
                logger.info("job function type {} not found i MO".format(job_pos_id))
            else:
                return_status[1] = True
                self._edit_klasse_title(mo_job_function_type["uuid"], title)
                logger.info("Updated job function type type: {}".format(job_pos_id))

        return "Job position updated"


@click.command()
@click.option(
    "--job-pos-id", type=click.STRING, help="Synchronize the provided job identifier."
)
@click.option(
    "--title",
    type=click.STRING,
    help="Title override, only has effect if job-pos-id is given.",
)
@click.option(
    "--sync-all", is_flag=True, type=click.BOOL, help="Synchronize all job identifiers."
)
def sync_jobid(job_pos_id, title, sync_all):
    """Job Position Synchronize tool."""
    setup_logging()

    if job_pos_id is None and sync_all is None:
        raise click.ClickException("Either job-pos-id or sync-all must be given")
    if job_pos_id and sync_all:
        raise click.ClickException("job-pos-id and sync-all are mutually exclusive")

    sync_tool = JobIdSync()

    if job_pos_id:
        print(job_pos_id)
        if title:
            print(sync_tool.sync_manually(job_pos_id, title))
        else:
            print(sync_tool.sync_from_sd(job_pos_id))

    if sync_all:
        sync_tool.sync_all_from_sd()
    logger.info("*Sync ended*")


if __name__ == "__main__":
    sync_jobid()



import json
from copy import deepcopy
from typing import Any
from typing import Dict
from unittest.mock import patch
from uuid import uuid4

import pytest
from parameterized import parameterized
from pydantic import BaseSettings
from pydantic import ValidationError

from sdlon.config import ChangedAtSettings
from sdlon.config import gen_json_file_settings_func
from sdlon.config import get_changed_at_settings
from sdlon.config import get_importer_settings
from sdlon.config import ImporterSettings

importer_json_file_settings = gen_json_file_settings_func(ImporterSettings)

DEFAULT_MOCK_SETTINGS = {
    "integrations.SD_Lon.employment_field": "extension_1",
    "integrations.SD_Lon.global_from_date": "2022-01-09",
    "integrations.SD_Lon.import.run_db": "run_db.sqlite",
    "integrations.SD_Lon.institution_identifier": "XYZ",
    "integrations.SD_Lon.job_function": "EmploymentName",
    "integrations.SD_Lon.monthly_hourly_divide": 50000,
    "integrations.SD_Lon.sd_user": "user",
    "integrations.SD_Lon.sd_password": "password",
    "municipality.code": 740,
    "municipality.name": "Kolding Kommune",
}

DEFAULT_EXPECTED_SETTINGS: Dict[str, Any] = {
    "sd_employment_field": "extension_1",
    "sd_global_from_date": "2022-01-09",
    "sd_import_run_db": "run_db.sqlite",
    "sd_institution_identifier": "XYZ",
    "sd_job_function": "EmploymentName",
    "sd_no_salary_minimum_id": None,
    "sd_monthly_hourly_divide": 50000,
    "sd_user": "user",
    "sd_password": "**********",
    "municipality_code": 740,
    "municipality_name": "Kolding Kommune",
    "mora_base": "http://mo-service:5000",
    "mox_base": "http://mox-service:8080",
    "sd_import_too_deep": [],
    "sd_importer_create_associations": True,
    "sd_importer_create_email_addresses": True,
    "sd_importer_employment_date_as_engagement_start_date": False,
    "sd_skip_job_functions": [],
    "sd_skip_employment_types": [],
    "sd_terminate_engagement_with_to_only": True,
    "sd_use_ad_integration": True,
    "cache_folder_path": "/opt/dipex/os2mo-data-import-and-export/tmp/",
    "cpr_uuid_map_path": "/opt/dipex/os2mo-data-import-and-export/settings/cpr_uuid_map.csv",
}

DEFAULT_FILTERED_JSON_SETTINGS = {
    "sd_employment_field": "extension_1",
    "sd_global_from_date": "2022-01-09",
    "sd_import_run_db": "run_db.sqlite",
    "sd_institution_identifier": "XYZ",
    "sd_job_function": "EmploymentName",
    "sd_monthly_hourly_divide": 50000,
    "sd_user": "user",
    "sd_password": "password",
    "municipality_code": 740,
    "municipality_name": "Kolding Kommune",
}

DEFAULT_CHANGED_AT_SETTINGS = {
    "sd_employment_field": "extension_1",
    "sd_global_from_date": "2022-01-09",
    "sd_import_run_db": "run_db.sqlite",
    "sd_institution_identifier": "XY",
    "sd_job_function": "EmploymentName",
    "sd_monthly_hourly_divide": 9000,
    "sd_password": "secret",
    "sd_user": "user",
}


@pytest.fixture
def mock_env_and_json(monkeypatch):
    monkeypatch.setenv("SD_USER", "env_user")
    monkeypatch.setattr("sdlon.config.load_settings", lambda: DEFAULT_MOCK_SETTINGS)


@pytest.fixture
def mock_env(monkeypatch):
    monkeypatch.setenv("SD_INSTITUTION_IDENTIFIER", "institution_id")
    monkeypatch.setenv("SD_USER", "env_user")
    monkeypatch.setenv("SD_PASSWORD", "env_pwd")
    monkeypatch.setenv("SD_JOB_FUNCTION", "EmploymentName")
    monkeypatch.setenv("SD_MONTHLY_HOURLY_DIVIDE", "80000")
    monkeypatch.setenv("SD_IMPORT_RUN_DB", "env_run_db")
    monkeypatch.setenv("SD_GLOBAL_FROM_DATE", "2022-01-09")


@patch("sdlon.config.load_settings")
def test_json_file_settings(mock_load_settings):
    # Arrange
    mock_load_settings.return_value = DEFAULT_MOCK_SETTINGS

    # Act
    settings = importer_json_file_settings(BaseSettings())

    # Assert
    assert settings == DEFAULT_FILTERED_JSON_SETTINGS


@patch("sdlon.config.load_settings")
def test_json_file_settings_remove_unknown_settings(mock_load_settings):
    # Arrange
    mock_settings = deepcopy(DEFAULT_MOCK_SETTINGS)
    mock_settings.update({"unknown": "property"})
    mock_load_settings.return_value = mock_settings

    # Act
    settings = importer_json_file_settings(BaseSettings())

    # Assert
    assert settings == DEFAULT_FILTERED_JSON_SETTINGS


@patch("sdlon.config.load_settings")
def test_empty_dict_on_file_not_found_error(mock_load_settings):
    # Arrange
    mock_load_settings.side_effect = FileNotFoundError()

    # Act
    json_settings = importer_json_file_settings(BaseSettings())

    # Assert
    assert json_settings == dict()


@patch("sdlon.config.load_settings")
def test_set_defaults(mock_load_settings):
    # Arrange
    mock_load_settings.return_value = DEFAULT_MOCK_SETTINGS

    # Act
    get_importer_settings.cache_clear()
    settings_input = get_importer_settings()

    # Assert
    assert json.loads(settings_input.json()) == DEFAULT_EXPECTED_SETTINGS


def test_forbid_extra_settings():
    with pytest.raises(ValidationError):
        ImporterSettings(
            municipality_name="name",
            municipality_code=100,
            sd_global_from_date="1970-01-01",
            sd_employment_field="extension_1",
            sd_import_run_db="run_db.sqlite",
            sd_institution_identifier="XY",
            sd_job_function="EmploymentName",
            sd_monthly_hourly_divide=9000,
            sd_password="secret",
            sd_user="user",
            forbidden="property",
        )


def test_env_settings_takes_precedence(mock_env_and_json):
    # Act
    get_importer_settings.cache_clear()
    settings = get_importer_settings()

    # Assert
    assert settings.sd_user == "env_user"


def test_pydantic_settings_set_correctly_when_json_settings_not_found(mock_env):
    # Act
    get_changed_at_settings.cache_clear()
    with patch("sdlon.config.load_settings") as mock_load_settings:
        mock_load_settings.side_effect = FileNotFoundError()
        settings = get_changed_at_settings()

    # Assert
    assert settings.sd_institution_identifier == "institution_id"
    assert settings.sd_user == "env_user"
    assert settings.sd_password.get_secret_value() == "env_pwd"
    assert settings.sd_job_function == "EmploymentName"
    assert settings.sd_monthly_hourly_divide == 80000
    assert settings.sd_import_run_db == "env_run_db"


@patch("sdlon.config.load_settings")
def test_override_default(mock_load_settings):
    # Arrange
    mock_settings = deepcopy(DEFAULT_MOCK_SETTINGS)
    mock_settings.update({"integrations.SD_Lon.sd_importer.create_associations": False})
    mock_load_settings.return_value = mock_settings

    # Act
    get_importer_settings.cache_clear()
    settings = get_importer_settings()

    # Assert
    assert not settings.sd_importer_create_associations


@parameterized.expand(
    [
        ("mora_base", "Not a URL"),
        ("mox_base", "Not a URL"),
        ("municipality_code", 98),
        ("municipality_code", 1000),
        ("sd_employment_field", "extension_"),
        ("sd_employment_field", "Invalid string"),
        ("sd_global_from_date", "Invalid string"),
        ("sd_job_function", "not allowed"),
        ("sd_monthly_hourly_divide", -1),
    ]
)
def test_special_values(key, value):
    # Arrange
    mock_settings = deepcopy(DEFAULT_MOCK_SETTINGS)
    mock_settings[key] = value

    # Act and assert
    with pytest.raises(ValidationError):
        ImporterSettings.parse_obj(mock_settings)


@parameterized.expand(["JobPositionIdentifier", "EmploymentName"])
def test_job_function_enums_allowed(job_function):
    assert ImporterSettings(
        municipality_name="name",
        municipality_code=100,
        sd_global_from_date="1970-01-01",
        sd_employment_field="extension_1",
        sd_import_run_db="run_db.sqlite",
        sd_institution_identifier="XY",
        sd_job_function=job_function,
        sd_monthly_hourly_divide=9000,
        sd_password="secret",
        sd_user="user",
    )


@parameterized.expand(
    [
        ("sd_fix_departments_root", str(uuid4()), "not a UUID"),
        ("sd_cprs", ["1234561234", "6543214321"], ["Not CPR"]),
        ("sd_cprs", ["0000000000"], "Not list of CPRs"),
        ("sd_exclude_cprs_mode", False, "Not a boolean"),
    ]
)
def test_changed_at_settings(key, valid_value, invalid_value):
    settings = deepcopy(DEFAULT_CHANGED_AT_SETTINGS)

    # The setting is optional or has a default
    assert ChangedAtSettings.parse_obj(settings)

    # ... and can be set to a valid_value
    settings.update({key: valid_value})
    assert ChangedAtSettings.parse_obj(settings)

    # ... but not an invalid_value
    settings.update({key: invalid_value})
    with pytest.raises(ValidationError):
        ChangedAtSettings.parse_obj(settings)



from copy import deepcopy
from csv import DictReader
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Tuple
from uuid import UUID

from more_itertools import flatten
from pydantic import BaseModel, Field

from os2mo_data_import.Clients.LoRa.model import Klasse
from os2mo_data_import.Clients.MO.model import Address, Employee, Engagement, \
    EngagementAssociation, OrgUnit
from os2mo_data_import.Clients.MO.model_parts.interface import MoObj
from os2mo_data_import.util import generate_uuid


class OrstedBusinessError(Exception):
    pass


class UUIDGenerator:
    def __init__(self, uuid_gen_seed: str = ""):
        self.__seed = uuid_gen_seed
        self.__cache: Dict[str, str] = {}  # Dict[user_key: uuid]

    def get_all(self) -> Dict[str, str]:
        return deepcopy(self.__cache)

    def get_uuid(self, key: str) -> str:
        """
        Generate uuid if missing
        :param key:
        :return:
        """
        if key in self.__cache:
            return self.__cache[key]

        uuid = generate_uuid(self.__seed + key)
        self.__cache[key] = uuid
        return uuid


class AddressKlasses(BaseModel):
    EMAIL: Klasse
    PHONE: Klasse
    MOBILE: Klasse
    HR_COUNTRY_CODE: Klasse
    CONSULTANT_EMAIL: Klasse
    HR_TDB_MOBILE_PHONE: Klasse
    HR_TDB_FIXED_PHONE: Klasse
    HR_LOCATION_ID_and_OFFICE: Klasse

    COUNTRY_OF_RESIDENCE: Klasse

    @classmethod
    def from_simplified_fields(
        cls,
        organisation_uuid: UUID,
        engagement_address_type_uuid: UUID,
        employee_address_type_uuid: UUID,
    ):
        gen_klass = partial(gen_single_klass, organisation_uuid=organisation_uuid)
        return AddressKlasses(
            EMAIL=gen_klass(
                facet_uuid=engagement_address_type_uuid,
                scope="EMAIL",
                title="EMAIL",
            ),
            PHONE=gen_klass(
                facet_uuid=engagement_address_type_uuid,
                scope="PHONE",
                title="PHONE",
            ),
            MOBILE=gen_klass(
                facet_uuid=engagement_address_type_uuid,
                scope="PHONE",
                title="MOBILE",
            ),
            HR_COUNTRY_CODE=gen_klass(
                facet_uuid=engagement_address_type_uuid,
                scope="TEXT",
                title="HR_COUNTRY_CODE",
            ),
            CONSULTANT_EMAIL=gen_klass(
                facet_uuid=engagement_address_type_uuid,
                scope="EMAIL",
                title="CONSULTANT_EMAIL",
            ),
            HR_TDB_MOBILE_PHONE=gen_klass(
                facet_uuid=engagement_address_type_uuid,
                scope="PHONE",
                title="HR_TDB_MOBILE_PHONE",
            ),
            HR_TDB_FIXED_PHONE=gen_klass(
                facet_uuid=engagement_address_type_uuid,
                scope="EMAIL",
                title="EMAIL",
            ),
            HR_LOCATION_ID_and_OFFICE=gen_klass(
                facet_uuid=engagement_address_type_uuid,
                scope="MULTIFIELD_TEXT",
                title="HR_LOCATION_ID_and_OFFICE",
            ),
            COUNTRY_OF_RESIDENCE=gen_klass(
                facet_uuid=employee_address_type_uuid,
                scope="TEXT",
                title="COUNTRY_OF_RESIDENCE",
            ),
        )


def gen_single_klass(
    facet_uuid: UUID, organisation_uuid: UUID, title: str, scope: Optional[str] = None
) -> Klasse:
    return Klasse.from_simplified_fields(
        facet_uuid=facet_uuid,
        uuid=generate_uuid("klasse" + str(facet_uuid) + title),
        user_key=title,
        title=title,
        scope=scope,
        organisation_uuid=organisation_uuid,
    )


class RawEmdData(BaseModel):
    Initials: str = Field(alias="Initials")
    First_Name: str = Field(alias="First Name")
    Last_Name: str = Field(alias="Last Name")
    Title: str = Field(alias="Title")
    Email: str = Field(alias="Email")
    Phone: str = Field(alias="Phone")
    Mobile: str = Field(alias="Mobile")
    Office: str = Field(alias="Office")
    Employment_Date: str = Field(alias="Employment Date")
    Departure_Date: str = Field(alias="Departure Date")
    Hr_Employee_No: str = Field(alias="Hr Employee No")
    Hr_Department_No: str = Field(alias="Hr Department No")
    Phonebook_Relevant: str = Field(alias="Phonebook Relevant")
    Hr_Employee_Type: str = Field(alias="Hr Employee Type")
    Hr_Location_Id: str = Field(alias="Hr Location Id")
    Hr_Country_Code: str = Field(alias="Hr Country Code")
    Hr_Cost_Center: str = Field(alias="Hr Cost Center")
    Hr_Hirakey_Relevant: str = Field(alias="Hr Hirakey Relevant")
    Hr_Active_Status: str = Field(alias="Hr Active Status")
    Hr_Legal_Company: str = Field(alias="Hr Legal Company")
    Consultant_Email: str = Field(alias="Consultant Email")
    Hr_Tdb_Mobile_Phone: str = Field(alias="Hr Tdb Mobile Phone")
    Hr_Tdb_Fixed_Phone: str = Field(alias="Hr Tdb Fixed Phone")
    Hr_Consultant_Type: str = Field(alias="Hr Consultant Type")
    Hr_Company_Code: str = Field(alias="Hr Company Code")
    Ee_Subgroup: str = Field(alias="Ee Subgroup")
    Country_Of_Residence: str = Field(alias="Country Of Residence")
    Seniority: str = Field(alias="Seniority")
    Career_Level: str = Field(alias="Career Level")
    Position_Level: str = Field(alias="Position Level")
    Career_Track: str = Field(alias="Career Track")
    Sender_Cost_Center: str = Field(alias="Sender Cost Center")

    def __business_validate(self):
        if self.Hr_Employee_Type and self.Hr_Consultant_Type:
            raise OrstedBusinessError(
                f"mutually exclusive: {self.Hr_Employee_Type}, "
                f"{self.Hr_Consultant_Type}"
            )
        if self.Office and not self.Hr_Location_Id:
            raise OrstedBusinessError(
                f"need location of office: {self.Office}, {self.Hr_Location_Id}"
            )
        if bool(self.Hr_Legal_Company) != bool(self.Hr_Company_Code):
            raise OrstedBusinessError(
                f"need both or neither: {self.Hr_Legal_Company}, {self.Hr_Company_Code}"
            )
        try:
            datetime.strptime(self.Employment_Date, "%d-%b-%y").strftime("%Y-%m-%d")
        except Exception:
            raise OrstedBusinessError(
                f"invalid employment_date format: {self.Employment_Date}"
            )

    def __generate_employee(self) -> Employee:
        name = f"{self.First_Name} {self.Last_Name}"
        person_uuid = generate_uuid(name)
        return Employee(uuid=person_uuid, name=name, seniority=self.Seniority)

    def __get_from_to_dates(self) -> Tuple[str, Optional[str]]:
        from_date = datetime.strptime(self.Employment_Date, "%d-%b-%y").strftime(
            "%Y-%m-%d"
        )
        to_date = None
        if self.Departure_Date:
            to_date = datetime.strptime(self.Departure_Date, "%d-%b-%y")
            # seems like Orsted data is weird, quickfix: roll ahead one centry
            if to_date == datetime(year=1999, month=12, day=31):
                to_date = datetime(year=2099, month=12, day=31)
            to_date = to_date.strftime("%Y-%m-%d")
        return from_date, to_date

    def __generate_engagement(
        self,
        *,
        person_uuid: UUID,
        job_function_uuid_generator: UUIDGenerator,
        engagement_type_uuid_generator: UUIDGenerator,
        org_unit_uuids: Dict[str, str],
        primary_uuid: UUID,
    ) -> Engagement:
        engagement_type = (
            self.Hr_Employee_Type
            if self.Hr_Employee_Type
            else self.Hr_Consultant_Type
        )

        job_function_uuid = job_function_uuid_generator.get_uuid(self.Title)
        engagement_type_uuid = engagement_type_uuid_generator.get_uuid(
            engagement_type
        )

        engagement_seed = "unique_engagement_salt_" + self.Initials
        engagement_uuid = generate_uuid(engagement_seed)

        from_date, to_date = self.__get_from_to_dates()

        return Engagement.from_simplified_fields(
            uuid=engagement_uuid,
            org_unit_uuid=org_unit_uuids[str(self.Hr_Department_No)],
            person_uuid=person_uuid,
            job_function_uuid=job_function_uuid,
            engagement_type_uuid=engagement_type_uuid,
            from_date=from_date,
            to_date=to_date,
            primary_uuid=primary_uuid,
            user_key=self.Initials,
            extension_1=self.Hr_Employee_No,
            extension_2=self.Hr_Hirakey_Relevant,
            extension_3=self.Hr_Active_Status,
            extension_4=self.Ee_Subgroup,
            extension_5=self.Career_Level,
            extension_6=self.Position_Level,
            extension_7=self.Career_Track,
        )

    def __generate_addresses(
        self,
        visible_uuid,
        not_visible_uuid,
        engagement_uuid,
        address_klasses: AddressKlasses,
        from_date,
        to_date,
        person_uuid,
        org_uuid: UUID,
    ) -> List[Address]:
        visibility_uuid = (
            visible_uuid if self.Phonebook_Relevant else not_visible_uuid
        )

        address_candidates = [
            (self.Email, None, address_klasses.EMAIL),
            (self.Phone, None, address_klasses.PHONE),
            (self.Mobile, None, address_klasses.MOBILE),
            (self.Hr_Country_Code, None, address_klasses.HR_COUNTRY_CODE),
            (self.Hr_Tdb_Fixed_Phone, None, address_klasses.HR_TDB_FIXED_PHONE),
            (self.Hr_Tdb_Mobile_Phone, None, address_klasses.HR_TDB_MOBILE_PHONE),
            (self.Consultant_Email, None, address_klasses.CONSULTANT_EMAIL),
            (
                self.Hr_Location_Id,
                self.Office,
                address_klasses.HR_LOCATION_ID_and_OFFICE,
            ),
        ]

        def to_address(
            value: str, value2: Optional[str], address_type_klasse: Klasse
        ) -> Optional[Address]:
            if not value:
                return
            if address_type_klasse.attributes.properties[0].scope == "PHONE":
                try:
                    int(value)
                except Exception:
                    return None
            return Address.from_simplified_fields(
                uuid=generate_uuid(
                    str(engagement_uuid) + str(address_type_klasse.uuid) + value + str(value2)
                ),
                value=value,
                value2=value2,
                address_type_uuid=address_type_klasse.uuid,
                engagement_uuid=engagement_uuid,
                from_date=from_date,
                to_date=to_date,
                visibility_uuid=visibility_uuid,
                org_uuid=org_uuid,
            )

        addresses = list(
            filter(
                lambda tmp: tmp is not None,
                [
                    to_address(value, value2, address_type_klasse)
                    for value, value2, address_type_klasse in address_candidates
                ],
            )
        )
        if self.Country_Of_Residence:
            addresses.append(
                Address.from_simplified_fields(
                    uuid=generate_uuid(
                        str(person_uuid)
                        + str(address_klasses.COUNTRY_OF_RESIDENCE.uuid)
                        + self.Country_Of_Residence
                    ),
                    value=self.Country_Of_Residence,
                    value2=None,
                    address_type_uuid=address_klasses.COUNTRY_OF_RESIDENCE.uuid,
                    person_uuid=person_uuid,
                    from_date=from_date,
                    to_date=to_date,
                    visibility_uuid=visibility_uuid,
                    org_uuid=org_uuid,
                )
            )

        return addresses

    def __generate_org_units(
        self,
        org_unit_type_uuid,
        org_unit_level_uuid,
        financial_org_unit_uuid,
        legal_org_unit_uuid,
    ) -> Tuple[Optional[OrgUnit], Optional[OrgUnit], Optional[OrgUnit]]:
        cost_center = None
        if self.Hr_Cost_Center:
            cost_center = OrgUnit.from_simplified_fields(
                uuid=generate_uuid("unique_org_unit_seed" + self.Hr_Cost_Center),
                user_key=self.Hr_Cost_Center,
                name=self.Hr_Cost_Center,
                parent_uuid=financial_org_unit_uuid,
                org_unit_hierarchy_uuid=None,
                org_unit_type_uuid=org_unit_type_uuid,
                org_unit_level_uuid=org_unit_level_uuid,
            )

        sender_cost_center = None
        if self.Sender_Cost_Center:
            sender_cost_center = OrgUnit.from_simplified_fields(
                uuid=generate_uuid("unique_org_unit_seed" + self.Sender_Cost_Center),
                user_key=self.Sender_Cost_Center,
                name=self.Sender_Cost_Center,
                parent_uuid=financial_org_unit_uuid,
                org_unit_hierarchy_uuid=None,
                org_unit_type_uuid=org_unit_type_uuid,
                org_unit_level_uuid=org_unit_level_uuid,
            )

        legal_company = None
        if self.Hr_Legal_Company:
            legal_company = OrgUnit.from_simplified_fields(
                uuid=generate_uuid("unique_org_unit_seed" + self.Hr_Legal_Company),
                user_key=self.Hr_Company_Code,
                name=self.Hr_Legal_Company,
                parent_uuid=legal_org_unit_uuid,
                org_unit_hierarchy_uuid=None,
                org_unit_type_uuid=org_unit_type_uuid,
                org_unit_level_uuid=org_unit_level_uuid,
            )
        return cost_center, sender_cost_center, legal_company

    @staticmethod
    def __generate_engagement_association(
        org_unit_uuids: Iterable[UUID],
        engagement_uuid: UUID,
        association_type_uuid: UUID,
        from_date: str,
        to_date: Optional[str],
    ) -> List[EngagementAssociation]:
        return [
            EngagementAssociation.from_simplified_fields(
                uuid=generate_uuid(
                    "engagement_association"
                    + str(engagement_uuid)
                    + str(org_unit_uuid)
                    + str(association_type_uuid)
                ),
                org_unit_uuid=org_unit_uuid,
                engagement_uuid=engagement_uuid,
                engagement_association_type_uuid=association_type_uuid,
                from_date=from_date,
                to_date=to_date,
            )
            for org_unit_uuid in org_unit_uuids
        ]

    def to_mo_objs(
        self,
        org_unit_uuids: Dict[str, str],
        primary_uuid: UUID,
        engagement_type_uuid_generator: UUIDGenerator,
        job_function_uuid_generator: UUIDGenerator,
        visible_uuid: UUID,
        not_visible_uuid: UUID,
        address_klasses: AddressKlasses,
        org_unit_type_uuid: UUID,
        financial_org_unit_uuid: UUID,
        legal_org_unit_uuid: UUID,
        engagement_association_type_uuid: UUID,
        org_unit_level_uuid: UUID,
        org_uuid: UUID,
    ):
        """
        :param org_unit_uuids: Hr_Department_No: uuid
        :param primary_uuid: Precreated Klasse -- I guess NOT primary for now?
        :param engagement_type_uuid_generator: Generate on-the-fly
        :param job_function_uuid_generator: Generate on-the-fly
        :param visible_uuid: Precreated Klasse
        :param not_visible_uuid: Precreated Klasse
        :param address_klasses: Obj containing precreated Address-related Klass'es
        :param org_unit_type_uuid: Precreated Klasse
        :param financial_org_unit_uuid: Precreated OU
        :param legal_org_unit_uuid: Precreated OU
        :param org_unit_level_uuid: Precreated Klasse
        :param engagement_association_type_uuid: Precreated Klasse
        :param org_uuid:
        :return:
        """

        try:
            self.__business_validate()
        except OrstedBusinessError:
            # traceback.print_exc()
            return

        employee = self.__generate_employee()

        engagement = self.__generate_engagement(
            person_uuid=employee.uuid,
            job_function_uuid_generator=job_function_uuid_generator,
            engagement_type_uuid_generator=engagement_type_uuid_generator,
            org_unit_uuids=org_unit_uuids,
            primary_uuid=primary_uuid,
        )

        addresses = self.__generate_addresses(
            visible_uuid=visible_uuid,
            not_visible_uuid=not_visible_uuid,
            engagement_uuid=engagement.uuid,
            address_klasses=address_klasses,
            from_date=engagement.validity.from_date,
            to_date=engagement.validity.to_date,
            person_uuid=employee.uuid,
            org_uuid=org_uuid,
        )

        org_units = list(
            filter(
                lambda ou: ou is not None,
                self.__generate_org_units(
                    org_unit_type_uuid=org_unit_type_uuid,
                    financial_org_unit_uuid=financial_org_unit_uuid,
                    legal_org_unit_uuid=legal_org_unit_uuid,
                    org_unit_level_uuid=org_unit_level_uuid,
                ),
            )
        )
        engagement_associations = self.__generate_engagement_association(
            org_unit_uuids=[ou.uuid for ou in org_units],
            engagement_uuid=engagement.uuid,
            association_type_uuid=engagement_association_type_uuid,
            from_date=engagement.validity.from_date,
            to_date=engagement.validity.to_date,
        )

        # order matters!
        return employee, org_units, engagement, engagement_associations, addresses


def read_emd(
    path: Path,
    org_unit_uuids: Dict[str, str],
    primary_uuid: UUID,
    engagement_type_uuid_generator: UUIDGenerator,
    job_function_uuid_generator: UUIDGenerator,
    visible_uuid: UUID,
    not_visible_uuid: UUID,
    address_klasses: AddressKlasses,
    org_unit_type_uuid: UUID,
    financial_org_unit_uuid: UUID,
    legal_org_unit_uuid: UUID,
    engagement_association_type_uuid: UUID,
    org_unit_level_uuid: UUID,
    org_uuid: UUID,
) -> Tuple[Iterable[Iterable[MoObj]], ...]:
    with path.open("r") as file:
        raw_emd_datas: Iterable[RawEmdData] = map(
            RawEmdData.parse_obj, DictReader(file, delimiter=";")
        )
        rows = [
            row.to_mo_objs(
                org_unit_uuids=org_unit_uuids,
                primary_uuid=primary_uuid,
                engagement_type_uuid_generator=engagement_type_uuid_generator,
                job_function_uuid_generator=job_function_uuid_generator,
                visible_uuid=visible_uuid,
                not_visible_uuid=not_visible_uuid,
                address_klasses=address_klasses,
                org_unit_type_uuid=org_unit_type_uuid,
                financial_org_unit_uuid=financial_org_unit_uuid,
                legal_org_unit_uuid=legal_org_unit_uuid,
                engagement_association_type_uuid=engagement_association_type_uuid,
                org_unit_level_uuid=org_unit_level_uuid,
                org_uuid=org_uuid,
            )
            for row in raw_emd_datas
        ]
        rows = filter(lambda objs: objs is not None, rows)
        # convert from rows to columns
        (
            employees,
            org_unitss,
            engagements,
            engagement_associationss,
            addressess,
        ) = zip(*rows)
        flat_ous = list(set(flatten(org_unitss)))  # get unique
        return (  # FIX THIS
            [[emp] for emp in employees],  # employees one at a time
            [[ou] for ou in flat_ous],  # one at a time for some reason
            [[eng] for eng in engagements],  # one at a time for some reason
            [[ea] for ea in flatten(engagement_associationss)],
            [[addr] for addr in flatten(addressess)]
            # one at a time for some reason
        )


if __name__ == "__main__":

    with Path("/home/mw/gir/gir_from_emd.csv").open("r") as f:
        raw_emd_dat = map(RawEmdData.parse_obj, DictReader(f, delimiter=";"))
        for x in raw_emd_dat:
            print(x)
            exit()



from datetime import datetime
from operator import itemgetter

import los_files
import pytest
import util
from aiohttp import ClientResponseError
from hypothesis import given
from hypothesis import strategies as st
from more_itertools import one

from .helpers import mock_config


class TestParseFilenames:
    def test_parses_prefix_correctly(self):
        filenames = [
            "xxxxOrg_nye_20210128_123742.csv",
            "Org_nye_20210128_123742.csv",
            "garbagebagasdasdads",
        ]

        expected_filename = "Org_nye_20210128_123742.csv"

        actual = los_files.parse_filenames(filenames, "Org_nye", datetime.min)

        actual_filename, _ = one(actual)

        assert expected_filename == actual_filename

    def test_filters_dates_correctly(self):
        filenames = [
            "Org_nye_20200128_123742.csv",
            "Org_nye_20201028_123742.csv",
            "Org_nye_20210128_123742.csv",
        ]

        expected = [
            ("Org_nye_20201028_123742.csv", datetime(2020, 10, 28, 12, 37, 42)),
            ("Org_nye_20210128_123742.csv", datetime(2021, 1, 28, 12, 37, 42)),
        ]

        actual = los_files.parse_filenames(
            filenames, "Org_nye", datetime(2020, 6, 1, 0, 0, 0)
        )

        assert expected == actual

    @given(
        st.permutations(
            [
                "Org_nye_20210101_000000.csv",
                "Org_nye_20200101_000000.csv",
                "Org_nye_20190101_000000.csv",
                "Org_nye_20180101_000000.csv",
            ]
        )
    )
    def test_sorts_output(self, filenames):
        expected_dates = [
            datetime(2018, 1, 1),
            datetime(2019, 1, 1),
            datetime(2020, 1, 1),
            datetime(2021, 1, 1),
        ]

        actual = los_files.parse_filenames(filenames, "Org_nye", datetime.min)
        actual_dates = list(map(itemgetter(1), actual))

        assert expected_dates == actual_dates


@pytest.mark.asyncio
async def test_terminate_details_handles_404_response(aioresponses):
    async def _run_test(ignored_http_statuses):
        mora_base = "http://example.com:8080"

        # Mock a 404 response from MO "terminate" API
        aioresponses.post(
            f"{mora_base}/service/details/terminate?force=1",
            status=404,
        )

        # Run `terminate_details`
        with mock_config(
            mora_base=mora_base,
            max_concurrent_requests=1,
            os2mo_chunk_size=1,
        ):
            async with util.get_client_session() as client_session:
                detail_payloads = [{"foo": "bar"}]
                await util.terminate_details(
                    client_session,
                    detail_payloads,
                    ignored_http_statuses=ignored_http_statuses,
                )

    # Expect no exceptions if 404 is ignored
    await _run_test(ignored_http_statuses=(404,))

    # Expect an exception if `raise_for_status` is called in `submit_payloads`
    with pytest.raises(ClientResponseError):
        await _run_test(ignored_http_statuses=None)



from collections.abc import Mapping
from functools import wraps


class AttrDict(dict):
    """Enable dot.notation access for a dict object.

    Example:
        script_result = AttrDict({"exit_code": 0})
        self.assertEqual(script_result.exit_code, 0)
    """

    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__  # type: ignore
    __delattr__ = dict.__delitem__  # type: ignore


def recursive_dict_update(original, updates):
    """Recursively update 'original' with keys from 'updates'.

    Example:
        original = {'alfa': {'beta': 2, 'charlie': 3}},
        updates = {'alfa': {'beta': 4}}
        # Non recursive update
        updated = {**original, **updates}
        self.assertEqual(updated, {'alfa': {'beta': 4}})
        # Recursive update
        r_updated = recursive_dict_update(original, updates)
        self.assertEqual(r_updated, {'alfa': {'beta': 4, 'charlie': 3}})

    Returns:
        dict: modified 'original'
    """
    for key, value in updates.items():
        if isinstance(value, Mapping):
            original[key] = recursive_dict_update(original.get(key, {}), value)
        else:
            original[key] = value
    return original


def dict_map(dicty, key_func=None, value_func=None, func=None):
    """Map the dict values.

    Example:
        input_dict = {1: 1, 2: 2, 3: 3}
        output_dict = dict_map(input_dict, value_func=lambda value: value ** 2)
        self.assertEqual(output_dict, {1: 1, 2: 4, 3: 6})

    Returns:
        dict: A dict where func has been applied to every value.
    """

    def identity(x):
        return x

    def tuple_identity(x, y):
        return (x, y)

    def help_call(func):
        def inner(x, **kwargs):
            try:
                return func(x, **kwargs)
            except TypeError:
                return func(x)

        return inner

    key_func = help_call(key_func or identity)
    value_func = help_call(value_func or identity)
    func = func or tuple_identity
    return dict(
        [
            func(key_func(key, value=value), value_func(value, key=key))
            for key, value in dicty.items()
        ]
    )


def dict_partition(func, dicty):
    """Partition the input dict into two using the predicate function.

    Example:
        input_dict = {0: 'value', 1: 'value': 2: 'value'}
        odd_dict, even_dict = dict_partition(
            lambda key, value: value % 2 == 0, input_dict
        )
        self.assertEqual(odd_dict, {1: 'value'})
        self.assertEqual(even_dict, {0: 'value', 2: 'value'})

    Returns:
        (dict, dict): A dict containing key-value pairs that failed the
                      predicate and a dict containing the key-value pairs
                      that passed the predicate.
    """
    falsy, truesy = {}, {}
    for key, value in dicty.items():
        write_dict = truesy if func(key, value) else falsy
        write_dict[key] = value
    return falsy, truesy


def dict_filter(func, dicty):
    return dict_partition(func, dicty)[1]


def dict_exclude(dicty, keys):
    return dict_filter(lambda key, value: key not in keys, dicty)


def dict_subset(dicty, keys):
    return dict_filter(lambda key, value: key in keys, dicty)


def duplicates(iterable):
    """Return set of duplicates from iterable.

    Example:
        input_list = [1, 5, 2, 4, 2, 1]
        dup_set = duplicates(input_list)
        self.assertEqual(dup_set, {1, 2})

    Returns:
        set: A set of the elements which are duplicates.
    """
    seen = set()
    return set(x for x in iterable if x in seen or seen.add(x))


def lower_list(listy):
    """Convert each element in the list to lower-case.

    Example:
        result = lower_list(['Alfa', 'BETA', 'gamma'])
        self.assertEqual(result, ['alfa', 'beta', 'gamma'])

    Args:
        listy: The list of strings to force into lowercase.

    Returns:
        list: A list where all contained the strings are lowercase.
    """
    return [x.lower() for x in listy]


def apply(func):
    """Return decorated / applied version of func.

    Example:
        @apply
        def tuple_expand(num, char):
            return char * num

        elements = [(1,'a'), (2, 'b'), (3, 'c')]
        elements = list(map(tuple_expand, elements))
        self.assertEqual(elements, ['a', 'bb', 'ccc'])

    Args:
        func: The function to apply arguments for

    Returns:
        wrapped function: Which has its one argument applied.
    """

    @wraps(func)
    def applied(tup):
        return func(*tup)

    return applied


def progress_iterator(elements, outputter=print, mod=10):
    """Output progress as iteration progresses.

    Example:
        elements = [(1,'a'), (2, 'b'), (3, 'c')]
        elements = progress_iterator(elements)
        elements = map(itemgetter(1), elements)
        self.assertEqual(list(elements), [1,2,3])

    Args:
        elements: list of elements,
        outputter: Function to call with progress strings,
        mod: The modulus for printing operations.

    Returns:
        Generator of objects in elements.
    """
    total = len(elements)
    for i, element in enumerate(elements, start=1):
        if i == 1 or i % mod == 0 or i == total:
            outputter("{}/{}".format(i, total))
        yield element



import os

from hypothesis import settings as h_settings
from hypothesis import Verbosity

h_settings.register_profile("ci", max_examples=100, deadline=None)
h_settings.register_profile("dev", max_examples=10)
h_settings.register_profile("debug", max_examples=10, verbosity=Verbosity.verbose)
h_settings.load_profile(os.getenv("HYPOTHESIS_PROFILE", "default"))



import os

from os2mo_helpers.mora_helpers import MoraHelper

from .ad_logger import start_logging

MORA_BASE = os.environ.get("MORA_BASE")


def find_mo_engagements_without_ad(helper):
    """ "
    Find a list of all current MO engagements that do not have a valid AD account
    registred as an IT-system.
    """
    org = helper.read_organisation()
    query = "o/{}/e?limit=1000000000"
    employees = helper._mo_lookup(org, query)

    active_users = set()
    missing_in_ad = set()
    i = 0
    for employee in employees["items"]:
        i += 1
        print("{}/{}".format(i, len(employees["items"])))
        uuid = employee["uuid"]
        engagements = helper.read_user_engagement(uuid)
        if engagements:
            active_users.add(uuid)

    for employee in active_users:
        ad_user = helper.get_e_username(employee, "Active Directory")
        # Notice: mora_helpers returns an empty string (not None) if no
        # account is found.
        if ad_user == "":
            missing_in_ad.add(employee)
    return missing_in_ad


if __name__ == "__main__":
    start_logging("ad_helpers.log")
    helper = MoraHelper(hostname=MORA_BASE, use_cache=False)

    missing_in_ad = find_mo_engagements_without_ad(helper)
    print(len(missing_in_ad))



import pytest
from glom import glom
from glom.core import PathAccessError
from hypothesis import given
from hypothesis import strategies as st

from ..read_ad_conf_settings import injected_settings
from ..read_ad_conf_settings import read_settings


def get_minimum_valid_settings(*extra_properties):
    properties = ["uuid_field", "org_unit_field"]
    for extra in filter(None, extra_properties):
        properties.append(extra)

    return {
        "integrations.ad": [
            {
                "system_user": "system_user",
                "password": "password",
                "search_base": "search_base",
                "cpr_field": "cpr_field",
                "properties": properties,
            }
        ],
        "integrations.ad.winrm_host": "winrm_host",
    }


def get_minimum_valid_writer_settings(level2orgunit_field=None):
    level2orgunit_field = level2orgunit_field or "l2org_field"
    settings = get_minimum_valid_settings(level2orgunit_field)
    settings.update(
        {
            "integrations.ad.write.level2orgunit_field": level2orgunit_field,
            "integrations.ad.write.level2orgunit_type": "l2org_type",
            "integrations.ad.write.upn_end": "upn_end",
            "integrations.ad.write.uuid_field": "uuid_field",
            "integrations.ad.write.org_unit_field": "org_unit_field",
        }
    )
    return settings


def test_minimum_valid_settings():
    read_settings(get_minimum_valid_settings())
    read_settings(get_minimum_valid_writer_settings())


def test_missing_winrm_setting():
    settings = get_minimum_valid_settings()
    del settings["integrations.ad.winrm_host"]
    with pytest.raises(Exception) as excinfo:
        read_settings(settings)
    assert "Missing hostname for remote management server" in str(excinfo.value)


@given(st.text(), st.text() | st.none())
def test_duplicated_field_names(ad_key, level2orgunit_field):
    settings = get_minimum_valid_writer_settings(level2orgunit_field)
    settings["integrations.ad_writer.mo_to_ad_fields"] = {"a": ad_key}
    settings["integrations.ad_writer.template_to_ad_fields"] = {ad_key: "c"}
    with pytest.raises(ValueError) as excinfo:
        read_settings(settings)
    expected = "Duplicated AD field names in settings: [%r]" % ad_key.lower()
    assert expected in str(excinfo.value)


@given(
    st.tuples(st.text(min_size=1), st.text() | st.none()).filter(lambda x: x[0] != x[1])
)
def test_missing_properties(tup):
    ad_key, level2orgunit_field = tup
    settings = get_minimum_valid_writer_settings(level2orgunit_field)
    settings["integrations.ad_writer.template_to_ad_fields"] = {ad_key: "c"}
    with pytest.raises(ValueError) as excinfo:
        read_settings(settings)
    expected = "Missing AD field names in properties: [%r]" % ad_key
    assert expected in str(excinfo.value)

    def test_injected_settings(self):
        title = "Stilling"
        old_settings = {
            "global": {
                "mora.base": "http://localhost:5000",
                "password": "pwd",
                "servers": [],
                "system_user": "os2mo",
                "winrm_host": "AD",
            },
            "primary": {
                "cpr_field": "EmployeeID",
                "method": "ntlm",
                "password": "pwd",
                "properties": [
                    "givenName",
                    "surname",
                    "distinguishedName",
                    "EmployeeID",
                    "title",
                ],
                "system_user": "os2mo",
            },
            "primary_write": {
                "cpr_field": "EmployeeID",
                "mo_to_ad_fields": {
                    "end_date": "info",
                    "unit_postal_code": "postalCode",
                },
                "org_field": "Department",
                "template_to_ad_fields": {"description": "{{" " sync_timestamp " "}}"},
            },
        }
        new_settings = {
            "ad_lifecycle_injected_settings": {
                "primary_write.mo_to_ad_fields.Title": title
            }
        }
        with pytest.raises(PathAccessError):
            assert (
                glom(old_settings, "primary_write.template_to_ad_fields.Title") != title
            )
        settings = injected_settings(
            "ad_lifecycle_injected_settings", old_settings, new_settings
        )

        assert glom(settings, "primary_write.mo_to_ad_fields.Title") == title



from unittest import TestCase
from unittest.mock import Mock
from unittest.mock import patch

from parameterized import parameterized

from ..ad_common import AD
from ..ad_exceptions import CommandFailure
from .mocks import MockAD


class TestRunPSScript(TestCase):
    def setUp(self):
        super().setUp()
        self._ad = MockAD()
        self._ps_script = "not actual PowerShell code"

    @parameterized.expand(
        [
            (b'{"foo": "bar}',),
            ('{"foo": "bar}',),
        ]
    )
    def test_invalid_json_raises_exception(self, invalid_json):
        with self._mock_run_ps_response(std_out=invalid_json):
            with self.assertRaises(ValueError) as ctx:
                self._ad._run_ps_script(self._ps_script)
                # Assert that exception message contains offending part of the
                # JSON doc, as well as the script that caused the error.
                msg = ctx.exception[0]
                self.assertIn(invalid_json, msg)
                self.assertIn(str(None), msg)

    def test_nonzero_status_code_raises_exception(self):
        with self._mock_run_ps_response(status_code=42):
            with self.assertRaises(CommandFailure):
                self._ad._run_ps_script(self._ps_script)

    def test_std_out_decoding(self):
        std_out = Mock()
        with self._mock_run_ps_response(std_out=std_out):
            # Passing a `Mock` instance causes this exception:
            # "TypeError: the JSON object must be str, bytes or bytearray, not
            # Mock". However, we just want to check the encoding passed to
            # `response.std_out.decode`, so we swallow the TypeError.
            with self.assertRaises(TypeError):
                self._ad._run_ps_script(self._ps_script)
                std_out.encode.assert_called_with(AD._encoding)

    def test_empty_std_out_returns_empty_dict(self):
        with self._mock_run_ps_response(std_out=None):
            empty_dict = self._ad._run_ps_script(self._ps_script)
            self.assertEqual(empty_dict, {})

    def _mock_run_ps_response(self, status_code=0, std_out=b"", std_err=b""):
        response = Mock()
        response.status_code = status_code
        response.std_out = std_out
        response.std_err = std_err
        return patch.object(self._ad.session, "run_ps", return_value=response)
